from typing import Union
from kgcnn.layers.casting import (
    CastBatchedIndicesToDisjoint, CastBatchedAttributesToDisjoint,
    CastDisjointToBatchedGraphState, CastDisjointToBatchedAttributes,
    CastBatchedGraphStateToDisjoint, CastRaggedAttributesToDisjoint,
    CastRaggedIndicesToDisjoint, CastDisjointToRaggedAttributes
)


template_cast_output_docs = r"""
    The standard model output template returns a single tensor of either "graph", "node", or "edge"
    embeddings specified by :obj:`output_embedding` within the model.
    The return tensor type is determined by :obj:`output_tensor_type` . Options are:

    graph:
        Tensor: Graph labels of either type
        
            - padded (Tensor): Batched tensor of graphs of shape `(batch, F)` .
            - disjoint (Tensor): Tensor with potential disjoint padded graphs of shape `(batch, F)` .
            - ragged (Tensor): Tensor of shape `(batch, F)` . No ragged dimension is needed here.

    nodes:
        Tensor: Node labels for the graph of either type:

            - ragged (RaggedTensor): Single tensor of shape `(batch, None, F)` .
            - padded (Tensor): Padded tensor of shape `(batch, N, F)` .
            - disjoint (Tensor): Disjoint representation of shape `([N], F)` .

    edges:
        Tensor: Edge labels for the graph of either type:

            - ragged (RaggedTensor): Single tensor of shape `(batch, None, F)` .
            - padded (Tensor): Padded tensor of shape `(batch, M, F)`
            - disjoint (Tensor): Disjoint representation of shape `([M], F)` .
"""


def template_cast_output(model_outputs,
                         output_embedding,
                         output_tensor_type,
                         input_tensor_type,
                         cast_disjoint_kwargs):
    """Template to cast graph, node or edge output to the desired tensor representation.

    Args:
        model_outputs (list): List of output and additional ID tensors. The list must always be
            [model_output, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges]
            but can have None in place of ID tensors if not required.
        output_embedding (str): Embedding of the graph output, either "graph", "node" or "edge".
        output_tensor_type (str): The tensor representation of model output.
        input_tensor_type (str): The tensor representation of model input.
        cast_disjoint_kwargs (dict): Kwargs for casting layers.

    Returns:
        Tensor: Keras output tensor.
    """
    out, batch_id_node, batch_id_edge, node_id, edge_id, count_nodes, count_edges = model_outputs

    # Output embedding choice
    if output_embedding == 'graph':
        if output_tensor_type in ["padded", "masked"]:
            # Here we could also modify the behaviour for direct disjoint without removing the padding via
            # remove_padded_disjoint_from_batched_output
            out = CastDisjointToBatchedGraphState(**cast_disjoint_kwargs)(out)
        elif output_tensor_type in ["ragged", "jagged"]:
            pass
        else:
            pass
    elif output_embedding == 'node':
        if output_tensor_type in ["padded", "masked"]:
            if "static_batched_node_output_shape" in cast_disjoint_kwargs:
                out_node_shape = cast_disjoint_kwargs["static_batched_node_output_shape"]
            else:
                out_node_shape = None
            out = CastDisjointToBatchedAttributes(static_output_shape=out_node_shape, **cast_disjoint_kwargs)(
                [out, batch_id_node, node_id, count_nodes])
        elif output_tensor_type in ["ragged", "jagged"]:
            out = CastDisjointToRaggedAttributes()([out, batch_id_node, node_id, count_nodes])
        else:
            pass
    elif output_embedding == 'edge':
        if output_tensor_type in ["padded", "masked"]:
            if "static_batched_edge_output_shape" in cast_disjoint_kwargs:
                out_edge_shape = cast_disjoint_kwargs["static_batched_edge_output_shape"]
            else:
                out_edge_shape = None
            out = CastDisjointToBatchedAttributes(static_output_shape=out_edge_shape, **cast_disjoint_kwargs)(
                [out, batch_id_edge, edge_id, count_edges])
        elif output_tensor_type in ["ragged", "jagged"]:
            out = CastDisjointToRaggedAttributes()([out, batch_id_edge, edge_id, count_edges])
        else:
            pass
    else:
        raise NotImplementedError("Unknown output embedding choice.")

    return out


template_cast_list_input_docs = r"""
    Template of listed graph input tensors, which should be compatible to previous kgcnn versions and
    defines the order as follows: :obj:`[nodes, edges, angles, edge_indices, angle_indices, graph_state, ...]` .
    Where '...' denotes further mask or ID tensors, which is required for certain input types (see below).
    Depending on the model, some inputs may not be used (see model description for information on supported inputs).
    For example if the model does not support angles and no graph attribute input, the input becomes:
    :obj:`[nodes, edges, edge_indices, ...]` .
    In case of crystal graphs lattice and translation information has to be added. This will give a possible input of
    :obj:`[nodes, edges, angles, edge_indices, angle_indices, graph_state, image_translation, lattice,...]` .
    Note that in place of nodes or edges also more than one tensor can be provided, depending on the model, for example
    :obj:`[nodes_1, nodes_2, edges_1, edges_2, edge_indices, ...]` .

    However, for future models we intend to used named inputs rather than a list that is sensible to ordering.
    Whether to use mask or length tensor for padded as well as further parameter of casting has to be set with
    (dict) :obj:`cast_disjoint_kwargs` .

    Padded or Masked Inputs:
        list: :obj:`[nodes, edges, angles, edge_indices, angle_indices, graph_state, image_translation, lattice,
        node_mask/node_count, edge_mask/edge_count, angle_mask/angle_count]`

            - nodes (Tensor): Node attributes of shape `(batch, N, F)` or `(batch, N)`
              using an embedding layer.
            - edges (Tensor): Edge attributes of shape `(batch, M, F)` or `(batch, M)`
              using an embedding layer.
            - angles (Tensor): Angle attributes of shape `(batch, M, F)` or `(batch, K)`
              using an embedding layer.
            - edge_indices (Tensor): Index list for edges of shape `(batch, M, 2)` referring to nodes.
            - angle_indices (Tensor): Index list for angles of shape `(batch, K, 2)` referring to edges.
            - graph_state (Tensor): Graph attributes of shape `(batch, F)` .
            - image_translation (Tensor): Indices of the periodic image the sending node is located in.
              Shape is `(batch, M, 3)` .
            - lattice (Tensor): Lattice matrix of the periodic structure of shape `(batch, 3, 3)` .
            - node_mask (Tensor): Mask for padded nodes of shape `(batch, N)` .
            - edge_mask (Tensor): Mask for padded edges of shape `(batch, M)` .
            - angle_mask (Tensor): Mask for padded angles of shape `(batch, K)` .
            - node_count (Tensor): Total number of nodes if padding is used of shape `(batch, )` .
            - edge_count (Tensor): Total number of edges if padding is used of shape `(batch, )` .
            - angle_count (Tensor): Total number of angle if padding is used of shape `(batch, )` .

    Ragged or Jagged Inputs:
        list: :obj:`[nodes, edges, angles, edge_indices, angle_indices, graph_state, image_translation, lattice]`

            - nodes (RaggedTensor): Node attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edges (RaggedTensor): Edge attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - angles (RaggedTensor): Angle attributes of shape `(batch, None, F)` or `(batch, None)`
              using an embedding layer.
            - edge_indices (RaggedTensor): Index list for edges of shape `(batch, None, 2)` referring to nodes.
            - angle_indices (RaggedTensor): Index list for angles of shape `(batch, None, 2)` referring to edges.
            - graph_state (Tensor): Graph attributes of shape `(batch, F)` .
            - image_translation (RaggedTensor): Indices of the periodic image the sending node is located in.
              Shape is `(batch, None, 3)` .
            - lattice (Tensor): Lattice matrix of the periodic structure of shape `(batch, 3, 3)` .

    Disjoint Input:
        list: :obj:`[nodes, edges, angles, edge_indices, angle_indices, graph_state, image_translation, lattice,
        graph_id_node, graph_id_edge, graph_id_angle, nodes_id, edges_id, angle_id, nodes_count, edges_count,
        angles_count]`

            - nodes (Tensor): Node attributes of shape `([N], F)` or `([N], )` using an embedding layer.
            - edges (Tensor): Edge attributes of shape `([M], F)` or `([M], )` using an embedding layer.
            - angles (Tensor): Angle attributes of shape `([K], F)` or `([K], )` using an embedding layer.
            - edge_indices (Tensor): Index list for edges of shape `(2, [M])` referring to nodes.
            - angle_indices (Tensor): Index list for angles of shape `(2, [K])` referring to edges.
            - graph_state (Tensor): Graph attributes of shape `(batch, F)` .
            - image_translation (Tensor): Indices of the periodic image the sending node is located in.
              Shape is `([M], 3)` .
            - lattice (Tensor): Lattice matrix of the periodic structure of shape `(batch, 3, 3)` .
            - graph_id_node (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
            - graph_id_edge (Tensor): ID tensor of batch assignment in disjoint graph of shape `([M], )` .
            - graph_id_angle (Tensor): ID tensor of batch assignment in disjoint graph of shape `([K], )` .
            - nodes_id (Tensor): The ID-tensor to assign each node to its respective graph of shape `([N], )` .
            - edges_id (Tensor): The ID-tensor to assign each edge to its respective graph of shape `([M], )` .
            - angle_id (Tensor): The ID-tensor to assign each edge to its respective graph of shape `([K], )` .
            - nodes_count (Tensor): Tensor of number of nodes for each graph of shape `(batch, )` .
            - edges_count (Tensor): Tensor of number of edges for each graph of shape `(batch, )` .
            - angles_count (Tensor): Tensor of number of angles for each graph of shape `(batch, )` .
"""


def template_cast_list_input(model_inputs: list,
                             input_tensor_type: str,
                             cast_disjoint_kwargs: dict,
                             mask_assignment: list = None,
                             index_assignment: list = None,
                             return_sub_id: bool = True):
    r"""Template to cast a list of model inputs to a list of disjoint tensors. The number of model inputs can be
    variable. The system is preset and explained by :obj:`template_cast_list_input_docs` .

    The ID information from the mask in tensor form is appended to the returned list.
    Ragged Tensor do not require a mask but use the mask information to generate joint ID tensors.

    Args:
        model_inputs (list): List of Keras inputs.
        input_tensor_type (str): Input tensor type. Either "padded", "ragged" or "disjoint".
        cast_disjoint_kwargs (dict): Kwargs for casting layers.
        mask_assignment (list): List that assigns Tensors to their mask.
            Inputs that do not require a mask must be marked by Nones. Different inputs can use the same mask.
        index_assignment (list): List that assigns index Tensors to their target to which the index refer to.
            Inputs that are not indices must be marked by Nones.
        return_sub_id (bool): Whether the returned list contains the sub-graph ID tensors.

    Returns:
        list: List of Keras Tensors for disjoint model.
    """
    out_tensor = []
    out_batch_id = []
    out_graph_id = []
    out_totals = []
    is_already_disjoint = False

    if input_tensor_type in ["padded", "masked"]:
        if mask_assignment is None or not isinstance(mask_assignment, (list, tuple)):
            raise ValueError("Mask assignment information is required or invalid.")

        reduced_mask = [x for x in mask_assignment if x is not None]
        if len(reduced_mask) == 0:
            num_mask = 0
        else:
            num_mask = max(reduced_mask) + 1
        if len(mask_assignment) + num_mask != len(model_inputs):
            raise ValueError()

        values_input = model_inputs[:-num_mask]
        mask_input = model_inputs[-num_mask:]

        if index_assignment is None:
            index_assignment = [None for _ in range(len(values_input))]
        if len(index_assignment) != len(mask_assignment):
            raise ValueError("Number of provided mask tensors does not match template specification.")

        out_tensor = [None for _ in range(len(values_input))]
        out_batch_id = [None for _ in range(num_mask)]
        out_graph_id = [None for _ in range(num_mask)]
        out_totals = [None for _ in range(num_mask)]

        for i, i_ref in enumerate(index_assignment):
            if i_ref is None:
                continue
            assert isinstance(i_ref, int), "Must provide positional index of for reference of indices."
            ref = values_input[i_ref]
            x = values_input[i]
            m, m_ref = mask_assignment[i], mask_assignment[i_ref]
            ref_mask = mask_input[m_ref]
            x_mask = mask_input[m]
            o_ref, o_x, b_r, b_x, g_r, g_x, t_r, t_x = CastBatchedIndicesToDisjoint(
                **cast_disjoint_kwargs)([ref, x, ref_mask, x_mask])
            out_tensor[i] = o_x
            # Important to not overwrite indices with simple values here!
            if out_tensor[i_ref] is None:
                out_tensor[i_ref] = o_ref
            out_batch_id[m] = b_x
            out_batch_id[m_ref] = b_r
            out_graph_id[m] = g_x
            out_graph_id[m_ref] = g_r
            out_totals[m] = t_x
            out_totals[m_ref] = t_r

        for i, x in enumerate(values_input):
            if out_tensor[i] is not None:
                continue
            m = mask_assignment[i]
            if m is None:
                out_tensor[i] = CastBatchedGraphStateToDisjoint(**cast_disjoint_kwargs)(x)
                continue

            x_mask = mask_input[m]
            o_x, bi, gi, tot = CastBatchedAttributesToDisjoint(**cast_disjoint_kwargs)([x, x_mask])
            out_tensor[i] = o_x
            out_batch_id[m] = bi
            out_graph_id[m] = gi
            out_totals[m] = tot

    elif input_tensor_type in ["ragged", "jagged"]:
        if index_assignment is None:
            index_assignment = [None for _ in range(len(model_inputs))]
        if len(index_assignment) != len(model_inputs):
            raise ValueError()

        reduced_mask = [x for x in mask_assignment if x is not None]
        if len(reduced_mask) == 0:
            num_mask = 0
        else:
            num_mask = max(reduced_mask) + 1

        out_tensor = [None for _ in range(len(model_inputs))]
        out_batch_id = [None for _ in range(num_mask)]
        out_graph_id = [None for _ in range(num_mask)]
        out_totals = [None for _ in range(num_mask)]

        for i, i_ref in enumerate(index_assignment):
            if i_ref is None:
                continue
            assert isinstance(i_ref, int), "Must provide positional index of for reference of indices."
            ref = model_inputs[i_ref]
            x = model_inputs[i]
            m, m_ref = mask_assignment[i], mask_assignment[i_ref]
            o_ref, o_x, b_r, b_x, g_r, g_x, t_r, t_x = CastRaggedIndicesToDisjoint(
                **cast_disjoint_kwargs)([ref, x])
            out_tensor[i] = o_x
            # Important to no overwrite indices with simple values here! This is the case if indices reference indices.
            if out_tensor[i_ref] is None:
                out_tensor[i_ref] = o_ref
            out_batch_id[m] = b_x
            out_batch_id[m_ref] = b_r
            out_graph_id[m] = g_x
            out_graph_id[m_ref] = g_r
            out_totals[m] = t_x
            out_totals[m_ref] = t_r

        for i, x in enumerate(model_inputs):
            if out_tensor[i] is not None:
                continue
            m = mask_assignment[i]
            if m is not None:
                o_x, bi, gi, tot = CastRaggedAttributesToDisjoint(**cast_disjoint_kwargs)(x)
                out_tensor[i] = o_x
                out_batch_id[m] = bi
                out_graph_id[m] = gi
                out_totals[m] = tot
            else:
                out_tensor[i] = x

    else:
        is_already_disjoint = True

    if is_already_disjoint:
        out = model_inputs
    else:
        out = out_tensor + out_batch_id
        if return_sub_id:
            out = out + out_graph_id
        out = out + out_totals

    return out


def template_cast_dict_input(model_inputs: dict,
                             input_tensor_type: str,
                             cast_disjoint_kwargs: dict,
                             mask_assignment: dict = None,
                             index_assignment: dict = None,
                             return_sub_id: bool = True,
                             rename_mask_to_id: dict = None):
    """Template to cast a dictionary of model inputs to a dict of disjoint tensors. The number of model inputs can be
    variable. The system is rather flexible and explained by :obj:`template_cast_dict_input_docs` .

    The ID information from the mask in tensor form is appended to the returned dictionary.
    Ragged Tensor do not require a mask but use the mask information to generate joint ID tensors.

    Args:
        model_inputs (dict): Dictionary of Keras inputs.
        input_tensor_type (str): Input tensor type. Either "padded", "ragged" or "disjoint".
        cast_disjoint_kwargs (dict): Kwargs for casting layers.
        mask_assignment (dict): Dictionary of mask name for each input that requires a mask and is cast
            to a disjoint tensor representation.
        index_assignment (dict): Dictionary of assigning indices to the name of their target tensors to which
            the indices refer to.
        return_sub_id (bool): Whether the returned dict contains the sub-graph ID tensors.
        rename_mask_to_id (dict): Mapping of mask names to ID names.

    Returns:
        dict: Model input tensors in disjoint representation.
    """

    is_already_disjoint = False
    out_tensor = {}
    out_batch_id = {}
    out_graph_id = {}
    out_totals = {}

    if input_tensor_type in ["padded", "masked"]:
        if mask_assignment is None or not isinstance(mask_assignment, dict):
            raise ValueError("Mask assignment information is required or invalid.")

        reduced_mask = list(set(list(mask_assignment.values())))

        values_input = {key: value for key, value in model_inputs.items() if key not in reduced_mask}
        mask_input = {key: value for key, value in model_inputs.items() if key in reduced_mask}

        if index_assignment is None:
            index_assignment = {}

        for name_index, name_ref in index_assignment.items():
            ref = values_input[name_ref]
            x = values_input[name_index]
            m, m_ref = mask_assignment[name_index], mask_assignment[name_ref]
            ref_mask = mask_input[m_ref]
            x_mask = mask_input[m]
            o_ref, o_x, b_r, b_x, g_r, g_x, t_r, t_x = CastBatchedIndicesToDisjoint(
                **cast_disjoint_kwargs)([ref, x, ref_mask, x_mask])
            out_tensor[name_index] = o_x
            # Important to not overwrite indices with simple values here! This is the case if indices reference indices.
            if name_ref not in out_tensor.keys():
                out_tensor[name_ref] = o_ref
            out_batch_id[m] = b_x
            out_batch_id[m_ref] = b_r
            out_graph_id[m] = g_x
            out_graph_id[m_ref] = g_r
            out_totals[m] = t_x
            out_totals[m_ref] = t_r

        for name_i, x in values_input.items():

            if name_i in out_tensor.keys():
                continue

            if name_i not in mask_assignment.keys():
                out_tensor[name_i] = CastBatchedGraphStateToDisjoint(**cast_disjoint_kwargs)(x)
                continue

            m = mask_assignment[name_i]
            x_mask = mask_input[m]
            o_x, bi, gi, tot = CastBatchedAttributesToDisjoint(**cast_disjoint_kwargs)([x, x_mask])
            out_tensor[name_i] = o_x
            out_batch_id[m] = bi
            out_graph_id[m] = gi
            out_totals[m] = tot

    elif input_tensor_type in ["ragged", "jagged"]:
        if mask_assignment is None or not isinstance(mask_assignment, dict):
            raise ValueError("Mask assignment information is required or invalid.")

        if index_assignment is None:
            index_assignment = {}

        reduced_mask = list(set(list(mask_assignment.values())))

        values_input = {key: value for key, value in model_inputs.items() if key not in reduced_mask}

        for name_index, name_ref in index_assignment.items():
            ref = values_input[name_ref]
            x = values_input[name_index]
            m, m_ref = mask_assignment[name_index], mask_assignment[name_ref]
            o_ref, o_x, b_r, b_x, g_r, g_x, t_r, t_x = CastRaggedIndicesToDisjoint(
                **cast_disjoint_kwargs)([ref, x])
            out_tensor[name_index] = o_x
            # Important to no overwrite indices with simple values here! This is the case if indices reference indices.
            if name_ref not in out_tensor.keys():
                out_tensor[name_ref] = o_ref
            out_batch_id[m] = b_x
            out_batch_id[m_ref] = b_r
            out_graph_id[m] = g_x
            out_graph_id[m_ref] = g_r
            out_totals[m] = t_x
            out_totals[m_ref] = t_r

        for name_i, x in values_input.items():
            if name_i in out_tensor.keys():
                continue

            if name_i not in mask_assignment.keys():
                out_tensor[name_i] = x
                continue

            m = mask_assignment[name_i]
            o_x, bi, gi, tot = CastRaggedAttributesToDisjoint(**cast_disjoint_kwargs)(x)
            out_tensor[name_i] = o_x
            out_batch_id[m] = bi
            out_graph_id[m] = gi
            out_totals[m] = tot

    else:
        is_already_disjoint = True

    # Rename IDs.
    def map_mask_key(k):
        mapping = {} if rename_mask_to_id is None else rename_mask_to_id
        if k in mapping.keys():
            return mapping[k]
        return k
    out_batch_id = {"graph_id_%s" % map_mask_key(key): value for key, value in out_batch_id.items()}
    out_graph_id = {"%s_id" % map_mask_key(key): value for key, value in out_graph_id.items()}
    out_totals = {"%s_count" % map_mask_key(key): value for key, value in out_totals.items()}

    if is_already_disjoint:
        out = model_inputs
    else:
        out = {}
        out.update(out_tensor)
        out.update(out_batch_id)
        if return_sub_id:
            out.update(out_graph_id)
        out.update(out_totals)

    return out



from pymatgen.core.structure import Structure
from typing import Optional
from networkx import MultiDiGraph
from .base import CrystalPreprocessor
from . import graph_builder


class RadiusAsymmetricUnitCell(CrystalPreprocessor):
    """Preprocessor that builds asymmetric unit graphs with radius-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords', 'multiplicity']
    edge_attributes = ['cell_translation', 'distance', 'symmop', 'offset']
    graph_attributes = ['lattice_matrix', 'spacegroup']

    def __init__(self, radius: float = 3.0, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            radius (float, optional): Cutoff radius for each atom in Angstrom units. Defaults to 3.0.
        """
        super(RadiusAsymmetricUnitCell, self).__init__(**kwargs)
        self.radius = radius

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=True)
        g = graph_builder.add_radius_bonds(g, radius=self.radius, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        g = graph_builder.to_asymmetric_unit_graph(g)
        return g


class KNNAsymmetricUnitCell(CrystalPreprocessor):
    """Preprocessor that builds asymmetric unit graphs with kNN-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords', 'multiplicity']
    edge_attributes = ['cell_translation', 'distance', 'symmop', 'offset']
    graph_attributes = ['lattice_matrix', 'spacegroup']

    def __init__(self, k: int = 12, tolerance: Optional[float] = 1e-9, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            k (int, optional): How many nearest neighbours to consider for edge selection. Defaults to 12.
            tolerance (Optional[float], optional): If tolerance is not None,
                edges with distances of the k-th nearest neighbor plus the tolerance value are included in the graph.
                Defaults to 1e-9.
        """
        super(KNNAsymmetricUnitCell, self).__init__(**kwargs)
        self.k = k
        self.tolerance = tolerance

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=True)
        g = graph_builder.add_knn_bonds(g, k=self.k, tolerance=self.tolerance, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        g = graph_builder.to_asymmetric_unit_graph(g)
        return g


class VoronoiAsymmetricUnitCell(CrystalPreprocessor):
    """Preprocessor that builds asymmetric unit graphs with Voronoi-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords', 'multiplicity']
    edge_attributes = ['cell_translation', 'distance', 'symmop', 'offset', 'voronoi_ridge_area']
    graph_attributes = ['lattice_matrix', 'spacegroup']

    def __init__(self, min_ridge_area: Optional[float] = 0.0, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            min_ridge_area (Optional[float], optional): Threshold value for ridge area between two Voronoi cells.
                If a ridge area between two voronoi cells is smaller than this value the corresponding edge between
                the atoms of the cells is excluded from the graph. Defaults to 0.0.
        """
        super(VoronoiAsymmetricUnitCell, self).__init__(**kwargs)
        self.min_ridge_area = min_ridge_area

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=True)
        g = graph_builder.add_voronoi_bonds(g, min_ridge_area=self.min_ridge_area, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        g = graph_builder.to_asymmetric_unit_graph(g)
        return g


class RadiusUnitCell(CrystalPreprocessor):
    """Preprocessor that builds unit cell graphs with radius-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset']
    graph_attributes = ['lattice_matrix']

    def __init__(self, radius: float = 3.0, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            radius (float, optional): Cutoff radius for each atom in Angstrom units. Defaults to 3.0.
        """
        super(RadiusUnitCell, self).__init__(**kwargs)
        self.radius = radius

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_radius_bonds(g, radius=self.radius, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        return g


class KNNUnitCell(CrystalPreprocessor):
    """Preprocessor that builds unit cell graphs with kNN-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset']
    graph_attributes = ['lattice_matrix']

    def __init__(self, k: int = 12, tolerance: Optional[float] = 1e-9, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            k (int, optional): How many nearest neighbours to consider for edge selection. Defaults to 12.
            tolerance (Optional[float], optional): If tolerance is not None,
                edges with distances of the k-th nearest neighbor plus the tolerance value are included in the graph.
                Defaults to 1e-9.
        """
        super(KNNUnitCell, self).__init__(**kwargs)
        self.k = k
        self.tolerance = tolerance

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_knn_bonds(g, k=self.k, tolerance=self.tolerance, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        return g


class VoronoiUnitCell(CrystalPreprocessor):
    """Preprocessor that builds unit cell graphs with Voronoi-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset', 'voronoi_ridge_area']
    graph_attributes = ['lattice_matrix']

    def __init__(self, min_ridge_area: Optional[float] = 0.0, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            min_ridge_area (Optional[float], optional): Threshold value for ridge area between two Voronoi cells.
                If a ridge area between two voronoi cells is smaller than this value the corresponding edge between
                the atoms of the cells is excluded from the graph. Defaults to 0.0.
        """
        super(VoronoiUnitCell, self).__init__(**kwargs)
        self.min_ridge_area = min_ridge_area

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_voronoi_bonds(g, min_ridge_area=self.min_ridge_area, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        return g


class RadiusSuperCell(CrystalPreprocessor):
    """Preprocessor that builds super-cell graphs with radius-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset']
    graph_attributes = ['lattice_matrix']

    def __init__(self, radius: float = 3.0, size: list = None, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            radius (float, optional): Cutoff radius for each atom in Angstrom units. Defaults to 3.0.
            size (list, optional): How many cells the crystal will get expanded into each dimension.
                Defaults to [3,3,3].
        """
        super(RadiusSuperCell, self).__init__(**kwargs)
        if size is None:
            size = [3, 3, 3]
        self.radius = radius
        self.size = size

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_radius_bonds(g, radius=self.radius, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        g = graph_builder.to_supercell_graph(g, size=self.size)
        return g


class KNNSuperCell(CrystalPreprocessor):
    """Preprocessor that builds super-cell graphs with kNN-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset']
    graph_attributes = ['lattice_matrix']

    def __init__(self, k: int = 12, tolerance: Optional[float] = 1e-9, size: list = None, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            k (int, optional): How many nearest neighbours to consider for edge selection. Defaults to 12.
            tolerance (Optional[float], optional): If tolerance is not None,
                edges with distances of the k-th nearest neighbor plus the tolerance value are included in the graph.
                Defaults to 1e-9.
            size (list, optional): How many cells the crystal will get expanded into each dimension.
                Defaults to [3,3,3].
        """
        super(KNNSuperCell, self).__init__(**kwargs)
        if size is None:
            size = [3, 3, 3]
        self.k = k
        self.tolerance = tolerance
        self.size = size

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_knn_bonds(g, k=self.k, tolerance=self.tolerance, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        g = graph_builder.to_supercell_graph(g, size=self.size)
        return g


class VoronoiSuperCell(CrystalPreprocessor):
    """Preprocessor that builds super-cell graphs with Voronoi-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset', 'voronoi_ridge_area']
    graph_attributes = ['lattice_matrix']

    def __init__(self, min_ridge_area: Optional[float] = 0.0, size: list = None, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            min_ridge_area (Optional[float], optional): Threshold value for ridge area between two Voronoi cells.
                If a ridge area between two voronoi cells is smaller than this value the corresponding edge between
                the atoms of the cells is excluded from the graph. Defaults to 0.0.
            size (list, optional): How many cells the crystal will get expanded into each dimension.
                Defaults to [3,3,3].
        """
        super(VoronoiSuperCell, self).__init__(**kwargs)
        if size is None:
            size = [3, 3, 3]
        self.size = size
        self.min_ridge_area = min_ridge_area

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_voronoi_bonds(g, min_ridge_area=self.min_ridge_area, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        g = graph_builder.to_supercell_graph(g, size=self.size)
        return g


class RadiusNonPeriodicUnitCell(CrystalPreprocessor):
    """Preprocessor that builds non-periodic graphs with radius-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset']
    graph_attributes = ['lattice_matrix']

    def __init__(self, radius: float = 3.0, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            radius (float, optional): Cutoff radius for each atom in Angstrom units. Defaults to 3.0.
        """
        super(RadiusNonPeriodicUnitCell, self).__init__(**kwargs)
        self.radius = radius

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_radius_bonds(g, radius=self.radius, inplace=True)
        g = graph_builder.to_non_periodic_unit_cell(g, add_reverse_edges=True, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        return g


class KNNNonPeriodicUnitCell(CrystalPreprocessor):
    """Preprocessor that builds non-periodic graphs with kNN-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset']
    graph_attributes = ['lattice_matrix']

    def __init__(self, k=12, tolerance=1e-9, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            k (int, optional): How many nearest neighbours to consider for edge selection. Defaults to 12.
            tolerance (float, optional): If tolerance is not None,
                edges with distances of the k-th nearest neighbor plus the tolerance value are included in the graph.
                Defaults to 1e-9.
        """
        super(KNNNonPeriodicUnitCell, self).__init__(**kwargs)
        self.k = k
        self.tolerance = tolerance

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_knn_bonds(g, k=self.k, tolerance=self.tolerance, inplace=True)
        g = graph_builder.to_non_periodic_unit_cell(g, add_reverse_edges=True, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        return g


class VoronoiNonPeriodicUnitCell(CrystalPreprocessor):
    """Preprocessor that builds non-periodic graphs with Voronoi-based edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = ['cell_translation', 'distance', 'offset', 'voronoi_ridge_area']
    graph_attributes = ['lattice_matrix']

    def __init__(self, min_ridge_area: Optional[float] = 0.0, **kwargs):
        """Initializes the crystal preprocessor.

        Args:
            min_ridge_area (Optional[float], optional): Threshold value for ridge area between two Voronoi cells.
                If a ridge area between two voronoi cells is smaller than this value the corresponding edge between
                the atoms of the cells is excluded from the graph. Defaults to 0.0.
        """
        super(VoronoiNonPeriodicUnitCell, self).__init__(**kwargs)
        self.min_ridge_area = min_ridge_area

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        g = graph_builder.add_voronoi_bonds(g, min_ridge_area=self.min_ridge_area, inplace=True)
        g = graph_builder.to_non_periodic_unit_cell(g, add_reverse_edges=True, inplace=True)
        g = graph_builder.add_edge_information(g, inplace=True)
        return g


class UnitCell(CrystalPreprocessor):
    """Preprocessor that builds unit cell graphs without any edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords']
    edge_attributes = []
    graph_attributes = ['lattice_matrix']

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=False)
        return g


class AsymmetricUnitCell(CrystalPreprocessor):
    """Preprocessor that builds asymmetric unit graphs without any edges for crystals."""

    node_attributes = ['atomic_number', 'frac_coords', 'coords', 'asymmetric_mapping', 'symmop', 'multiplicity']
    edge_attributes = []
    graph_attributes = ['lattice_matrix', 'spacegroup']

    def call(self, structure: Structure) -> MultiDiGraph:
        """Builds the crystal graph (networkx.MultiDiGraph) for the pymatgen structure.

        Args:
            structure (Structure): Structure to convert to a crystal graph.

        Returns:
            MultiDiGraph: Crystal graph for the provided crystal.
        """
        if isinstance(structure, MultiDiGraph):
            g = structure
        else:
            g = graph_builder.structure_to_empty_graph(structure, symmetrize=True)
        return g



import keras as ks
import keras.saving
from keras import ops
from typing import Union
from kgcnn.models.utils import get_model_class
from keras.saving import deserialize_keras_object, serialize_keras_object
from keras.backend import backend


# In keras 3.0.0 there is no `ops.gradient()` function yet.
# Backend specific gradient implementation in the following.
if backend() == "tensorflow":
    import tensorflow as tf
elif backend() == "torch":
    import torch
elif backend() == "jax":
    import jax.numpy as jnp
    import jax
    from functools import partial
else:
    raise NotImplementedError("Backend '%s' not supported for force model." % backend())


@ks.saving.register_keras_serializable(package='kgcnn', name='EnergyForceModel')
class EnergyForceModel(ks.models.Model):
    r"""Outer model to wrap a normal invariant GNN to predict forces from energy predictions via partial derivatives.
    The Force :math:`\vec{F_i}` on Atom :math:`i` is given by

    .. math::

        \vec{F_i} = - \vec{\nabla}_i E_{\text{total}}

    Note that the model simply returns the tensor type of the coordinate input for forces. No casting is done
    by this class. This means that the model returns a ragged, disjoint or padded tensor depending on the tensor
    type of the coordinates.
    """

    def __init__(self,
                 inputs: Union[dict, list] = None,
                 model_energy=None,
                 coordinate_input: Union[int, str] = 1,
                 output_as_dict: bool = True,
                 ragged_validate: bool = False,
                 output_to_tensor: bool = True,
                 output_squeeze_states: bool = False,
                 nested_model_config: bool = True,
                 is_physical_force: bool = True,
                 use_batch_jacobian: bool = None,
                 name: str = None,
                 outputs: Union[dict, list] = None
                 ):
        """Initialize Force model with an energy model.

        Args:
            inputs (list): List of inputs as dictionary kwargs of keras input layers.
            model_energy (ks.models.Model, dict): Keras model os deserialization dictionary for a keras model.
            coordinate_input (int): Position of the coordinate input.
            output_as_dict (bool, tuple): Names for the output if a dictionary should be returned. Or simply a bool
                which will use the names "energy" and "force".
            ragged_validate (bool): Whether to validate ragged or jagged tensors.
            output_to_tensor: Deprecated.
            output_squeeze_states (bool): Whether to squeeze state/energy dimension for forces
                in case of a single energy value.
            nested_model_config (bool): Whether there is a config for the energy model.
            is_physical_force (bool): Whether to return the physical force, e.g. the negative gradient of the energy.
            use_batch_jacobian: Deprecated.
            name (str): Name of the model.
            outputs: List of outputs as dictionary kwargs similar to inputs.
        """
        super().__init__()
        if model_energy is None:
            raise ValueError("Require valid model in `model_energy` for force prediction.")
        # Input for model_energy.
        self._model_energy = model_energy
        self.name = name
        if isinstance(model_energy, ks.models.Model):
            # Ignoring module_name and class_name.
            self.energy_model = model_energy
        elif isinstance(model_energy, dict):
            if "module_name" not in model_energy:
                self.energy_model = deserialize_keras_object(model_energy)
            else:
                self.energy_model_class = get_model_class(model_energy["module_name"], model_energy["class_name"])
                self.energy_model = self.energy_model_class(**model_energy["config"])
        else:
            raise TypeError("Input `model_energy` must be dict or `ks.models.Model` . Can not deserialize model.")

        # Additional parameters of io and behavior of this class.
        self.ragged_validate = ragged_validate
        self.coordinate_input = coordinate_input
        self.output_squeeze_states = output_squeeze_states
        self.is_physical_force = is_physical_force
        self.nested_model_config = nested_model_config
        self._force_outputs = outputs

        self.output_as_dict = output_as_dict
        if isinstance(output_as_dict, bool):
            self.output_as_dict_use = output_as_dict
            self.output_as_dict_names = ("energy", "force")
        elif isinstance(output_as_dict, (list, tuple)):
            self.output_as_dict_use = True
            self.output_as_dict_names = (output_as_dict[0], output_as_dict[1])
        else:
            self.output_as_dict_use = False

        energy_output_config = outputs[self.output_as_dict_names[0]] if self.output_as_dict_use else outputs[0]
        self._expected_energy_states = energy_output_config["shape"][0]

        # We can try to infer the model inputs from energy model, if not given explicit.
        self._inputs_to_force_model = inputs
        if self._inputs_to_force_model is None:
            if self.nested_model_config and isinstance(model_energy, dict):
                self._inputs_to_force_model = model_energy["config"]["inputs"]

        if backend() == "tensorflow":
            self._call_grad_backend = self._call_grad_tf
        elif backend() == "torch":
            self._call_grad_backend = self._call_grad_torch
        elif backend() == "jax":
            self._call_grad_backend = self._call_grad_jax
        else:
            raise NotImplementedError("Backend '%s' not supported for force model." % backend())

    def build(self, input_shape):
        self.energy_model.build(input_shape)
        self.built = True

    def _call_grad_tf(self, inputs, training=False, **kwargs):

        x_in = inputs[self.coordinate_input]
        with tf.GradientTape(persistent=True) as tape:
            if isinstance(x_in, tf.RaggedTensor):
                x, splits = x_in.values, x_in.row_splits
            else:
                x, splits = x_in, None
            tape.watch(x)
            eng = self.energy_model.call(inputs, training=training, **kwargs)
            eng_sum = tf.reduce_sum(eng, axis=0, keepdims=False)
            e_grad = [eng_sum[i] for i in range(eng_sum.shape[-1])]
        e_grad = [tf.expand_dims(tape.gradient(e_i, x), axis=-1) for e_i in e_grad]
        e_grad = tf.concat(e_grad, axis=-1)

        if self.output_squeeze_states:
            e_grad = tf.squeeze(e_grad, axis=-1)

        if isinstance(x_in, tf.RaggedTensor):
            e_grad = tf.RaggedTensor.from_row_splits(e_grad, splits, validate=self.ragged_validate)

        return eng, e_grad

    def _call_grad_torch(self, inputs, training=False, **kwargs):

        x = inputs[self.coordinate_input]
        with torch.enable_grad():
            x.requires_grad = True
            eng = self.energy_model.call(inputs, training=training, **kwargs)
            eng_sum = eng.sum(dim=0)
            e_grad = torch.cat([
                torch.unsqueeze(torch.autograd.grad(eng_sum[i], x, create_graph=True, allow_unused=True)[0], dim=-1) for i in
                range(eng.shape[-1])], dim=-1)

        if self.output_squeeze_states:
            e_grad = torch.squeeze(e_grad, dim=-1)
        return eng, e_grad

    def _call_grad_jax(self, inputs, training=False, **kwargs):

        @partial(jax.jit, static_argnames=['pos'])
        def energy_reduce(*inputs, pos: int = 0):
            eng_temp = self.energy_model.call(inputs, training=training, **kwargs)
            eng_sum = jnp.sum(eng_temp, axis=0)[pos]
            return eng_sum

        grad_fn = jax.grad(energy_reduce, argnums=self.coordinate_input)
        all_grad = [grad_fn(*inputs, pos=i) for i in range(self._expected_energy_states)]
        eng = self.energy_model.call(inputs, training=training, **kwargs)
        e_grad = jnp.concatenate([jnp.expand_dims(x, axis=-1) for x in all_grad], axis=-1)

        if self.output_squeeze_states:
            e_grad = jnp.squeeze(e_grad, axis=-1)
        return eng, e_grad

    def call(self, inputs, training=False, **kwargs):

        eng, e_grad = self._call_grad_backend(inputs, training=training, **kwargs)

        if self.is_physical_force:
            e_grad = -e_grad

        if self.output_as_dict_use:
            return {self.output_as_dict_names[0]: eng, self.output_as_dict_names[1]: e_grad}
        else:
            return eng, e_grad

    def get_config(self):
        """Get config."""
        # Keras model does not provide config from base class.
        # conf = super(EnergyForceModel, self).get_config()
        conf = {}
        # Serialize class if _model_energy is not dict.
        if isinstance(self._model_energy, dict):
            model_energy = self._model_energy
        else:
            model_energy = serialize_keras_object(self._model_energy)
        conf.update({
            "model_energy": model_energy,
            "coordinate_input": self.coordinate_input,
            "output_as_dict": self.output_as_dict,
            "ragged_validate": self.ragged_validate,
            # "output_to_tensor": self.output_to_tensor,
            "output_squeeze_states": self.output_squeeze_states,
            "nested_model_config": self.nested_model_config,
            # "use_batch_jacobian": self.use_batch_jacobian,
            "inputs": self._inputs_to_force_model,
            "outputs": self._force_outputs
        })
        return conf



from keras.layers import Dense, Add, Multiply, Activation
from kgcnn.layers.modules import Embedding
from kgcnn.layers.gather import GatherNodes
from kgcnn.layers.relational import RelationalDense
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.mlp import MLP, GraphMLP


def model_disjoint(
        inputs,
        use_node_embedding,
        input_node_embedding=None,
        depth=None,
        dense_modulation_kwargs=None,
        dense_relation_kwargs=None,
        activation_kwargs=None,
        output_embedding=None,
        output_mlp=None,
        node_pooling_kwargs=None
):
    n, edge_relations, edi, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)

    # Model
    for i in range(0, depth):
        n_i, n_j = GatherNodes(split_indices=[0, 1], concat_axis=None)([n, edi])
        # Note: This maybe could be done more efficiently.
        gamma = RelationalDense(**dense_modulation_kwargs)([n_i, edge_relations])
        beta = RelationalDense(**dense_modulation_kwargs)([n_i, edge_relations])
        h_j = RelationalDense(**dense_relation_kwargs)([n_j, edge_relations])
        m = Multiply()([h_j, gamma])
        m = Add()([m, beta])
        h = AggregateLocalEdges(pooling_method="sum")([n, m, edi])
        n = Activation(**activation_kwargs)(h)

    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes(**node_pooling_kwargs)([count_nodes, n, batch_id_node])  # will return tensor
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":  # Node labeling
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `GNNFilm`")

    return out



import numpy as np
from kgcnn.molecule.graph_rdkit import MolecularGraphRDKit
from kgcnn.graph.base import GraphPreProcessorBase
from kgcnn.molecule.methods import inverse_global_proton_dict
from kgcnn.molecule.io import parse_list_to_xyz_str
from kgcnn.molecule.encoder import OneHotEncoder
from kgcnn.utils.serial import serialize
from kgcnn.molecule.serial import deserialize_encoder

_mol_graph_interface = MolecularGraphRDKit


class SetMolBondIndices(GraphPreProcessorBase):
    r"""Preprocessor to compute chemical bonds from coordinates via a :obj:`MolGraphInterface` .

    Args:
        node_coordinates (str): Name of atomic coordinates array of shape `(N, 3)` .
        node_symbol (str): Name of atomic symbol as numpy array of shape `(N, )` .
        node_number (str): Name of atomic numbers array of shape `(N, )` .
        edge_indices (str): Name to assign edge indices to.
        edge_number (str): Name to assign the edge number/order to.
        name (str): Name of this preprocessor.
    """

    def __init__(self, *, node_coordinates: str = "node_coordinates", node_symbol: str = "node_symbol",
                 node_number: str = "node_number",
                 edge_indices: str = "edge_indices", edge_number: str = "edge_number",
                 name="set_mol_bond_indices", **kwargs):
        super().__init__(name=name, **kwargs)
        self._to_obtain.update({"node_coordinates": node_coordinates, "node_number": node_number,
                                "node_symbol": node_symbol})
        self._to_assign = [edge_indices, edge_number]
        self._config_kwargs.update({
            "edge_indices": edge_indices, "node_coordinates": node_coordinates, "node_number": node_number,
            "node_symbol": node_symbol, "edge_number": edge_number})

    def call(self, node_coordinates: np.ndarray, node_symbol: np.ndarray, node_number: np.ndarray):
        if node_symbol is None:
            node_symbol = [inverse_global_proton_dict(int(x)) for x in node_number]
        else:
            node_symbol = [str(x) for x in node_symbol]
        mol = _mol_graph_interface()
        mol = mol.from_xyz(parse_list_to_xyz_str([node_symbol, node_coordinates.tolist()], number_coordinates=3))
        if mol.mol is None:
            return None, None
        idx, edge_num = mol.edge_number
        return idx, edge_num


class SetMolAttributes(GraphPreProcessorBase):
    """Preprocessor to compute molecular attributes from graph arrays that make a valid molecule
    via a :obj:`MolGraphInterface` .
    See :obj:`MoleculeNetDataset` which uses a callbacks but has identical nomenclature.

    .. code-block:: python

        from kgcnn.data.datasets.QM7Dataset import QM7Dataset
        from kgcnn.molecule.preprocessor import SetMolAttributes
        ds = QM7Dataset()
        pp = SetMolAttributes()
        print(pp(ds[0]))

    Args:
        nodes (list): List of atomic properties for attributes.
        edges (list): List of bond properties for attributes.
        graph (list): List of molecular properties for attributes.
        encoder_nodes (dict): Dictionary of node attribute encoders.
        encoder_edges (dict): Dictionary of edge attribute encoders.
        encoder_graph (dict): Dictionary of graph attribute encoders.
        node_coordinates (str): Name of numpy array storing atomic coordinates.
        node_symbol (str): Name of numpy array storing atomic symbol.
        node_number (str): Name of numpy array storing atomic number.
        edge_indices (str): Name of numpy array storing atomic bond indices.
        edge_number (str): Name of numpy array storing atomic bond order.
        node_attributes (str): Name to assign node attributes to.
        edge_attributes (str): Name to assign edge attributes to.
        graph_attributes (str):  Name to assign graph attributes to.
        name (str): Name of the preprocessor.
    """

    _default_node_attributes = [
        'Symbol', 'TotalDegree', 'FormalCharge', 'NumRadicalElectrons', 'Hybridization',
        'IsAromatic', 'IsInRing', 'TotalNumHs', 'CIPCode', "ChiralityPossible", "ChiralTag"
    ]
    _default_node_encoders = {
        'Symbol': OneHotEncoder(
            ['B', 'C', 'N', 'O', 'F', 'Si', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Te', 'I', 'At'],
            dtype="str"
        ),
        'Hybridization': OneHotEncoder([2, 3, 4, 5, 6]),
        'TotalDegree': OneHotEncoder([0, 1, 2, 3, 4, 5], add_unknown=False),
        'TotalNumHs': OneHotEncoder([0, 1, 2, 3, 4], add_unknown=False),
        'CIPCode': OneHotEncoder(['R', 'S'], add_unknown=False, dtype='str'),
        "ChiralityPossible": OneHotEncoder(["1"], add_unknown=False, dtype='str'),
    }
    _default_edge_attributes = ['BondType', 'IsAromatic', 'IsConjugated', 'IsInRing', 'Stereo']
    _default_edge_encoders = {
        'BondType': OneHotEncoder([1, 2, 3, 12], add_unknown=False),
        'Stereo': OneHotEncoder([0, 1, 2, 3], add_unknown=False)
    }
    _default_graph_attributes = ['ExactMolWt', 'NumAtoms']
    _default_graph_encoders = {}

    def __init__(self, *,
                 nodes: list = None, edges: list = None, graph: list = None,
                 encoder_nodes: dict = None,
                 encoder_edges: dict = None,
                 encoder_graph: dict = None,
                 node_coordinates: str = "node_coordinates", node_symbol: str = "node_symbol",
                 node_number: str = "node_number",
                 edge_indices: str = "edge_indices", edge_number: str = "edge_number",
                 node_attributes: str = "node_attributes", edge_attributes: str = "edge_attributes",
                 graph_attributes: str = "graph_attributes",
                 name="set_mol_attributes", **kwargs):
        super().__init__(name=name, **kwargs)
        nodes = nodes if nodes is not None else self._default_node_attributes
        edges = edges if edges is not None else self._default_edge_attributes
        graph = graph if graph is not None else self._default_graph_attributes
        encoder_nodes = encoder_nodes if encoder_nodes is not None else self._default_node_encoders
        encoder_edges = encoder_edges if encoder_edges is not None else self._default_edge_encoders
        encoder_graph = encoder_graph if encoder_graph is not None else self._default_graph_encoders

        self._to_obtain.update({"node_coordinates": node_coordinates, "node_number": node_number,
                                "node_symbol": node_symbol, "edge_indices": edge_indices, "edge_number": edge_number})
        self._to_assign = [node_attributes, edge_attributes, graph_attributes, edge_indices, edge_number]
        self._call_kwargs = {
            "nodes": nodes,
            "edges": edges,
            "graph": graph,
            "encoder_nodes": {key: deserialize_encoder(value) for key, value in encoder_nodes.items()},
            "encoder_edges": {key: deserialize_encoder(value) for key, value in encoder_edges.items()},
            "encoder_graph": {key: deserialize_encoder(value) for key, value in encoder_graph.items()}
        }
        self._config_kwargs.update({
            "edge_indices": edge_indices, "node_coordinates": node_coordinates, "node_number": node_number,
            "node_symbol": node_symbol, "edge_number": edge_number,
            "node_attributes": node_attributes, "edge_attributes": edge_attributes,
            "graph_attributes": graph_attributes,
            "nodes": nodes,
            "edges": edges,
            "graph": graph,
            "encoder_nodes": {key: serialize(value) for key, value in encoder_nodes.items()},
            "encoder_edges": {key: serialize(value) for key, value in encoder_edges.items()},
            "encoder_graph": {key: serialize(value) for key, value in encoder_graph.items()}
        })

    def call(self,
             nodes: list, edges: list, graph: list,
             encoder_nodes: dict,
             encoder_edges: dict,
             encoder_graph: dict,
             node_coordinates: np.ndarray, node_symbol: np.ndarray, node_number: np.ndarray,
             edge_indices: np.ndarray, edge_number: np.ndarray):
        if node_symbol is None:
            node_symbol = [inverse_global_proton_dict(int(x)) for x in node_number]
        else:
            node_symbol = [str(x) for x in node_symbol]
        mol = _mol_graph_interface()
        mol.from_list(node_symbol, edge_indices, edge_number, conformer=node_coordinates)
        n_att = mol.node_attributes(nodes, encoder=encoder_nodes)
        _, e_att = mol.edge_attributes(edges, encoder=encoder_edges)
        g_att = mol.graph_attributes(graph, encoder=encoder_graph)
        idx, en = mol.edge_number
        return n_att, e_att, g_att, idx, en



from keras.layers import Dense
from kgcnn.layers.conv import SchNetInteraction
from kgcnn.layers.geom import NodePosition, NodeDistanceEuclidean, GaussBasisLayer, ShiftPeriodicLattice
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.pooling import PoolingNodes


def model_disjoint(
        inputs,
        use_node_embedding: bool = None,
        input_node_embedding: dict = None,
        make_distance: bool = None,
        expand_distance: bool = None,
        gauss_args: dict = None,
        interaction_args: dict = None,
        node_pooling_args: dict = None,
        depth: int = None,
        last_mlp: dict = None,
        output_embedding: str = None,
        use_output_mlp: bool = None,
        output_mlp: dict = None):
    n, x, disjoint_indices, batch_id_node, count_nodes = inputs

    # Optional Embedding.
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)

    if make_distance:
        pos1, pos2 = NodePosition()([x, disjoint_indices])
        ed = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ed = x

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Model
    n = Dense(interaction_args["units"], activation='linear')(n)
    for i in range(0, depth):
        n = SchNetInteraction(**interaction_args)([n, ed, disjoint_indices])

    n = GraphMLP(**last_mlp)([n, batch_id_node, count_nodes])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `SchNet` .")

    return out


def model_disjoint_crystal(
        inputs,
        use_node_embedding: bool = None,
        input_node_embedding: dict = None,
        make_distance: bool = None,
        expand_distance: bool = None,
        gauss_args: dict = None,
        interaction_args: dict = None,
        node_pooling_args: dict = None,
        depth: int = None,
        last_mlp: dict = None,
        output_embedding: str = None,
        use_output_mlp: bool = None,
        output_mlp: dict = None):
    n, x, disjoint_indices, edge_image, lattice, batch_id_node, batch_id_edge, count_nodes = inputs

    # Optional Embedding.
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)

    if make_distance:
        pos1, pos2 = NodePosition()([x, disjoint_indices])
        pos2 = ShiftPeriodicLattice()([pos2, edge_image, lattice, batch_id_edge])
        ed = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ed, _, _, _ = x

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Model
    n = Dense(interaction_args["units"], activation='linear')(n)
    for i in range(0, depth):
        n = SchNetInteraction(**interaction_args)([n, ed, disjoint_indices])

    n = GraphMLP(**last_mlp)([n, batch_id_node, count_nodes])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `SchNet` .")

    return out



from keras.layers import Flatten, Dense
from kgcnn.layers.mlp import MLP, GraphMLP
from kgcnn.layers.modules import Embedding, ZerosLike
from ._layers import HamNaiveDynMessage, HamNetFingerprintGenerator, HamNetNaiveUnion, HamNetGRUUnion


def model_disjoint(
        inputs,
        use_node_embedding,
        use_edge_embedding,
        input_node_embedding=None,
        input_edge_embedding=None,
        given_coordinates=None,
        gru_kwargs=None,
        message_kwargs=None,
        fingerprint_kwargs=None,
        output_embedding=None,
        output_mlp=None,
        union_type_edge=None,
        union_type_node=None,
        depth=None
):
    # Model implementation with disjoint representation.
    n, q_ftr, ed, edi, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    # Generate coordinates.
    if given_coordinates:
        p_ftr = ZerosLike()(q_ftr)
    else:
        # Use Hamiltonian engine to get p, q coordinates.
        raise NotImplementedError("Hamiltonian engine not yet implemented")

    # Initialization
    n = Dense(units=gru_kwargs["units"], activation="tanh")(n)
    ed = Dense(units=gru_kwargs["units"], activation="tanh")(ed)
    p = p_ftr
    q = q_ftr

    # Message passing.
    for i in range(depth):
        # Message step
        nu, eu = HamNaiveDynMessage(**message_kwargs)([n, ed, p, q, edi])

        # Node updates
        if union_type_node == "gru":
            n = HamNetGRUUnion(**gru_kwargs)([n, nu])
        elif union_type_node == "naive":
            n = HamNetNaiveUnion(units=gru_kwargs["units"])([n, nu])
        else:
            n = nu

        # Edge updates
        if union_type_edge == "gru":
            ed = HamNetGRUUnion(**gru_kwargs)([ed, eu])
        elif union_type_edge == "naive":
            ed = HamNetNaiveUnion(units=gru_kwargs["units"])([ed, eu])
        else:
            ed = eu

    # Fingerprint generator for graph embedding.
    if output_embedding == 'graph':
        out = HamNetFingerprintGenerator(**fingerprint_kwargs)([count_nodes, n, batch_id_node])
        out = Flatten()(out)  # will be tensor.
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for `HamNet` .")

    return out



import pandas as pd
from typing import Optional
import importlib.resources
import kgcnn.crystal.periodic_table as periodic_table_module

# CSV file is downloaded from:
# https://pubchem.ncbi.nlm.nih.gov/rest/pug/periodictable/CSV/?response_type=save&response_basename=PubChemElements_all

try:
    # >= Python 3.9
    periodic_table_csv = importlib.resources.files(periodic_table_module) / 'periodic_table.csv'
except:
    # < Python 3.9
    with importlib.resources.path(periodic_table_module, 'periodic_table.csv') as file_name:
        periodic_table_csv = file_name


class PeriodicTable:
    """Utility class to provide further element type information for crystal graph node embeddings."""
    
    def __init__(self, csv_path=periodic_table_csv,
                 normalize_atomic_mass=True,
                 normalize_atomic_radius=True,
                 normalize_electronegativity=True,
                 normalize_ionization_energy=True,
                 imputation_atomic_radius=209.46,  # mean value
                 imputation_electronegativity=1.18,  # educated guess (based on neighbour elements)
                 imputation_ionization_energy=8.):  # mean value
        self.data = pd.read_csv(csv_path)
        self.data['AtomicRadius'].fillna(imputation_atomic_radius, inplace=True)
        # Pm, Eu, Tb, Yb are inside the mp_e_form dataset, but have no electronegativity value
        self.data['Electronegativity'].fillna(imputation_electronegativity, inplace=True)
        self.data['IonizationEnergy'].fillna(imputation_ionization_energy, inplace=True)
        if normalize_atomic_mass:
            self._normalize_column('AtomicMass')
        if normalize_atomic_radius:
            self._normalize_column('AtomicRadius')
        if normalize_electronegativity:
            self._normalize_column('Electronegativity')
        if normalize_ionization_energy:
            self._normalize_column('IonizationEnergy')
            
    def _normalize_column(self, column):
        self.data[column] = (self.data[column] - self.data[column].mean()) / self.data[column].std()

    def get_symbol(self, z: Optional[int] = None):
        if z is None:
            return self.data['Symbol'].to_list()
        else:
            return self.data.loc[z-1]['Symbol']
    
    def get_atomic_mass(self, z: Optional[int] = None):
        if z is None:
            return self.data['AtomicMass'].to_list()
        else:
            return self.data.loc[z-1]['AtomicMass']
    
    def get_atomic_radius(self, z: Optional[int] = None):
        if z is None:
            return self.data['AtomicRadius'].to_list()
        else:
            return self.data.loc[z-1]['AtomicRadius']
    
    def get_electronegativity(self, z: Optional[int] = None):
        if z is None:
            return self.data['Electronegativity'].to_list()
        else:
            return self.data.loc[z-1]['Electronegativity']
    
    def get_ionization_energy(self, z: Optional[int] = None):
        if z is None:
            return self.data['IonizationEnergy'].to_list()
        else:
            return self.data.loc[z-1]['IonizationEnergy']

    def get_oxidation_states(self, z: Optional[int] = None):
        if z is None:
            return list(map(self.parse_oxidation_state_string, self.data['OxidationStates'].to_list()))
        else:
            oxidation_states = self.data.loc[z-1]['OxidationStates']
            return self.parse_oxidation_state_string(oxidation_states, encode=True)
    
    @staticmethod
    def parse_oxidation_state_string(s: str, encode: bool = True):
        if encode:
            oxidation_states = [0] * 14
            if isinstance(s, float):
                return oxidation_states
            for i in s.split(','):
                oxidation_states[int(i)-7] = 1
        else:
            oxidation_states = []
            if isinstance(s, float):
                return oxidation_states
            for i in s.split(','):
                oxidation_states.append(int(i))
        return oxidation_states



import keras as ks
import keras.saving
import keras.optimizers.schedules
# from keras_core import ops


@ks.saving.register_keras_serializable(package='kgcnn', name='LinearWarmupExponentialDecay')
class LinearWarmupExponentialDecay(ks.optimizers.schedules.LearningRateSchedule):
    r"""This schedule combines a linear warmup with an exponential decay.
    Combines :obj:` tf.optimizers.schedules.PolynomialDecay` with an actual increase during warmup
    and :obj:`tf.optimizers.schedules.ExponentialDecay` after.

    Introduced by `DimeNetPP <https://arxiv.org/abs/2011.14115>`__ .

    The closed-from learning rate schedule for learning rate :math:`\eta` for :math:`s_0` warmup and decay
    :math:`S_\tau` is given as a function of steps :math:`s` below (deduced from documentation of keras modules).

    .. math::

        \eta (s) = \eta_0 \;  \gamma ^ {s / S_\tau}  \; [1 - \frac{s_0-1}{s_0} \frac{s_0 - \text{min}(s_0, S)}{s_0}]

    This class has been updated to be compatible with `GemNet <https://arxiv.org/abs/2106.08903>`__ training.

    """

    def __init__(self, learning_rate, warmup_steps, decay_steps, decay_rate, staircase: bool = False):
        """Initialize class.

        Args:
            learning_rate: Learning rate to use.
            warmup_steps: Number of warmup steps.
            decay_steps: Number of which to decay the learning rate.
            decay_rate: Factor to reduce the learning rate.
            staircase (bool): If True use staircase decay and not (continuous) exponential decay.
        """
        super().__init__()
        self._input_config_settings = {"learning_rate": learning_rate, "warmup_steps": warmup_steps,
                                       "decay_steps": decay_steps, "decay_rate": decay_rate, "staircase": staircase}
        self.warmup = ks.optimizers.schedules.PolynomialDecay(
            1 / warmup_steps, warmup_steps, end_learning_rate=1)
        self.decay = ks.optimizers.schedules.ExponentialDecay(
            learning_rate, decay_steps, decay_rate, staircase=staircase)

    def __call__(self, step):
        """Decay learning rate as a functions of steps.

        Args:
            step: Current step of training.

        Returns:
            float: New learning rate.
        """
        return self.warmup(step) * self.decay(step)

    @property
    def initial_learning_rate(self):
        return self.decay.initial_learning_rate

    @initial_learning_rate.setter
    def initial_learning_rate(self, value):
        self.decay.initial_learning_rate = value

    def get_config(self):
        """Get config for this class."""
        config = {}
        config.update(self._input_config_settings)
        return config


@ks.saving.register_keras_serializable(package='kgcnn', name='KerasPolynomialDecaySchedule')
class KerasPolynomialDecaySchedule(ks.optimizers.schedules.PolynomialDecay):
    r"""This schedule inherits from :obj:` ks.optimizers.schedules.PolynomialDecay` ."""

    def __init__(self, dataset_size: int, batch_size: int, epochs: int, lr_start: float = 0.0005,
                 lr_stop: float = 1e-5):
        """Initialize class.

        Args:
            dataset_size (int): Size of the dataset.
            batch_size (int): Batch size for training.
            epochs (int): Total epochs to run schedule on.
            lr_start (int): Learning rate at the start.
            lr_stop (int): Final learning rate.
        """
        self._input_config_settings = {"lr_start": lr_start, "lr_stop": lr_stop,
                                       "epochs": epochs, "batch_size": batch_size, "dataset_size": dataset_size}
        steps_per_epoch = dataset_size / batch_size
        num_steps = epochs * steps_per_epoch
        super().__init__(initial_learning_rate=lr_start, decay_steps=num_steps, end_learning_rate=lr_stop)

    def get_config(self):
        """Get config for this class."""
        config = {}
        config.update(self._input_config_settings)
        return config



from ._make import make_model, model_default
from ._make import make_model_weighted, model_default_weighted

__all__ = [
    "make_model",
    "model_default",
    "make_model_weighted",
    "model_default_weighted"
]




from kgcnn.layers.mlp import MLP, GraphMLP, RelationalMLP
from keras.layers import Concatenate, Dense, Multiply, Add
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.norm import GraphBatchNormalization
from ._wacsf import wACSFRad, wACSFAng
from ._layers import CorrectPartialCharges
from ._acsf import ACSFG2, ACSFG4, ACSFConstNormalization


def model_disjoint_weighted(
        inputs,
        node_pooling_args: dict = None,
        w_acsf_ang_kwargs: dict = None,
        w_acsf_rad_kwargs: dict = None,
        normalize_kwargs: dict = None,
        const_normalize_kwargs: dict = None,
        mlp_kwargs: dict = None,
        output_embedding: str = None,
        use_output_mlp: bool = None,
        output_mlp: dict = None,
        predict_dipole: bool = None
):
    # Make input
    node_input, xyz_input, edge_index_input, angle_index_input, tot_charge, batch_id_node, count_nodes = inputs

    # ACSF representation.
    rep_rad = wACSFRad(**w_acsf_rad_kwargs)([node_input, xyz_input, edge_index_input])
    rep_ang = wACSFAng(**w_acsf_ang_kwargs)([node_input, xyz_input, angle_index_input])
    rep = Concatenate()([rep_rad, rep_ang])

    # Normalization
    if normalize_kwargs:
        rep = GraphBatchNormalization(**normalize_kwargs)([rep, batch_id_node, count_nodes])
    if const_normalize_kwargs:
        rep = ACSFConstNormalization(**const_normalize_kwargs)(rep)

    # learnable NN.
    n = RelationalMLP(**mlp_kwargs)([rep, node_input, batch_id_node, count_nodes])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
        if predict_dipole:
            pc = Dense(units=1)(n)
            tc = PoolingNodes(pooling_method="sum")([count_nodes, pc, batch_id_node])
            if tot_charge is not None:
                pc_correct = CorrectPartialCharges()([tc, tot_charge, count_nodes, batch_id_node])
                pc = Add()([pc, pc_correct])
            p_dip = Multiply()([pc, xyz_input])
            dip = PoolingNodes(pooling_method="sum")([count_nodes, p_dip, batch_id_node])
            out = [out, dip]
            if tot_charge is None:
                out = out + [tc]
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `HDNNP2nd` .")

    return out


def model_disjoint_behler(
        inputs,
        node_pooling_args: dict = None,
        normalize_kwargs: dict = None,
        const_normalize_kwargs: dict = None,
        g2_kwargs: dict = None,
        g4_kwargs: dict = None,
        mlp_kwargs: dict = None,
        output_embedding: str = None,
        use_output_mlp: bool = None,
        output_mlp: dict = None,
        predict_dipole: bool = None
):
    # Make input
    node_input, xyz_input, edge_index_input, angle_index_input, tot_charge, batch_id_node, count_nodes = inputs

    # ACSF representation.
    rep_g2 = ACSFG2(**ACSFG2.make_param_table(**g2_kwargs))([node_input, xyz_input, edge_index_input])
    rep_g4 = ACSFG4(**ACSFG4.make_param_table(**g4_kwargs))([node_input, xyz_input, angle_index_input])
    rep = Concatenate()([rep_g2, rep_g4])

    # Normalization
    if normalize_kwargs:
        rep = GraphBatchNormalization(**normalize_kwargs)([rep, batch_id_node, count_nodes])
    if const_normalize_kwargs:
        rep = ACSFConstNormalization(**const_normalize_kwargs)(rep)

    # learnable NN.
    n = RelationalMLP(**mlp_kwargs)([rep, node_input, batch_id_node, count_nodes])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
        if predict_dipole:
            pc = Dense(units=1)(n)
            tc = PoolingNodes(pooling_method="sum")([count_nodes, pc, batch_id_node])
            if tot_charge is not None:
                pc_correct = CorrectPartialCharges()([tc, tot_charge, count_nodes, batch_id_node])
                pc = Add()([pc, pc_correct])
            p_dip = Multiply()([pc, xyz_input])
            dip = PoolingNodes(pooling_method="sum")([count_nodes, p_dip, batch_id_node])
            out = [out, dip]
            if tot_charge is None:
                out = out + [tc]
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `HDNNP2nd`")

    return out


def model_disjoint_atom_wise(
        inputs,
        node_pooling_args: dict = None,
        mlp_kwargs: dict = None,
        output_embedding: str = None,
        use_output_mlp: bool = None,
        output_mlp: dict = None,
        predict_dipole: bool = None
):
    # Make input
    node_input, rep_input, tot_charge, batch_id_node, count_nodes = inputs

    # learnable NN.
    n = RelationalMLP(**mlp_kwargs)([rep_input, node_input, batch_id_node, count_nodes])

    # Output embedding choice
    if output_embedding == 'graph':
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])
        if use_output_mlp:
            out = MLP(**output_mlp)(out)
        if predict_dipole:
            pc = Dense(units=1)(n)
            tc = PoolingNodes(pooling_method="sum")([count_nodes, pc, batch_id_node])
            if tot_charge is not None:
                pc_correct = CorrectPartialCharges()([tc, tot_charge, count_nodes, batch_id_node])
                pc = Add()([pc, pc_correct])
            p_dip = Multiply()([pc, xyz_input])
            dip = PoolingNodes(pooling_method="sum")([count_nodes, p_dip, batch_id_node])
            out = [out, dip]
            if tot_charge is None:
                out = out + [tc]
    elif output_embedding == 'node':
        out = n
        if use_output_mlp:
            out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `HDNNP2nd`")

    return out



from keras.layers import Layer, GRU
from kgcnn.layers.casting import CastDisjointToBatchedAttributes


class PoolingNodesGRU(Layer):

    def __init__(self, units, static_output_shape=None,
                 activation='tanh', recurrent_activation='sigmoid',
                 use_bias=True, kernel_initializer='glorot_uniform',
                 recurrent_initializer='orthogonal',
                 bias_initializer='zeros', kernel_regularizer=None,
                 recurrent_regularizer=None, bias_regularizer=None, kernel_constraint=None,
                 recurrent_constraint=None, bias_constraint=None, dropout=0.0,
                 recurrent_dropout=0.0, reset_after=True, seed=None,
                 **kwargs):
        super(PoolingNodesGRU, self).__init__(**kwargs)
        self.units = units
        self.static_output_shape = static_output_shape
        self.cast_layer = CastDisjointToBatchedAttributes(
            static_output_shape=static_output_shape, return_mask=True)
        self.gru = GRU(
            units=units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            reset_after=reset_after,
            seed=seed
        )

    def call(self, inputs, **kwargs):
        n, mask = self.cast_layer(inputs)
        out = self.gru(n, mask=mask)
        return out

    def get_config(self):
        config = super(PoolingNodesGRU, self).get_config()
        config.update({"units": self.units, "static_output_shape": self.static_output_shape})
        conf_gru = self.gru.get_config()
        param_list = ["units", "activation", "recurrent_activation",
                      "use_bias", "kernel_initializer",
                      "recurrent_initializer",
                      "bias_initializer", "kernel_regularizer",
                      "recurrent_regularizer", "bias_regularizer", "kernel_constraint",
                      "recurrent_constraint", "bias_constraint", "dropout",
                      "recurrent_dropout", "reset_after"]
        for x in param_list:
            if x in conf_gru.keys():
                config.update({x: conf_gru[x]})
        return config



import keras as ks
from keras.layers import Dense, Add, Layer


class GRUUpdate(Layer):
    r"""Gated recurrent unit for updating node or edge embeddings.
    As proposed by `NMPNN <http://arxiv.org/abs/1704.01212>`__ .
    """

    def __init__(self, units,
                 activation='tanh', recurrent_activation='sigmoid',
                 use_bias=True, kernel_initializer='glorot_uniform',
                 recurrent_initializer='orthogonal',
                 bias_initializer='zeros', kernel_regularizer=None,
                 recurrent_regularizer=None, bias_regularizer=None, kernel_constraint=None,
                 recurrent_constraint=None, bias_constraint=None, dropout=0.0,
                 recurrent_dropout=0.0, reset_after=True,
                 **kwargs):
        r"""Initialize layer.

        Args:
            units (int): Units for GRU.
            activation: Activation function to use. Default: hyperbolic tangent
                (`tanh`). If you pass None, no activation is applied
                (ie. "linear" activation: `a(x) = x`).
            recurrent_activation: Activation function to use for the recurrent step.
                Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
                applied (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean, (default `True`), whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix,
                used for the linear transformation of the inputs. Default:
                `glorot_uniform`.
            recurrent_initializer: Initializer for the `recurrent_kernel`
                weights matrix, used for the linear transformation of the recurrent state.
                Default: `orthogonal`.
            bias_initializer: Initializer for the bias vector. Default: `zeros`.
            kernel_regularizer: Regularizer function applied to the `kernel` weights
                matrix. Default: `None`.
            recurrent_regularizer: Regularizer function applied to the
                `recurrent_kernel` weights matrix. Default: `None`.
            bias_regularizer: Regularizer function applied to the bias vector. Default:
                `None`.
            kernel_constraint: Constraint function applied to the `kernel` weights
                matrix. Default: `None`.
            recurrent_constraint: Constraint function applied to the `recurrent_kernel`
                weights matrix. Default: `None`.
            bias_constraint: Constraint function applied to the bias vector. Default:
                `None`.
            dropout: Float between 0 and 1. Fraction of the units to drop for the
                linear transformation of the inputs. Default: 0.
            recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
                the linear transformation of the recurrent state. Default: 0.
            reset_after: GRU convention (whether to apply reset gate after or
                before matrix multiplication). False = "before",
                True = "after" (default and CuDNN compatible).
        """
        super(GRUUpdate, self).__init__(**kwargs)
        self.units = units

        self.gru_cell = ks.layers.GRUCell(
            units=units,
            activation=activation, recurrent_activation=recurrent_activation,
            use_bias=use_bias, kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout, reset_after=reset_after
        )

    def build(self, input_shape):
        """Build layer."""
        super(GRUUpdate, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [nodes, updates]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - updates (Tensor): Matching node updates of shape ([N], F)
            mask: Mask for inputs. Default is None.

        Returns:
           Tensor: Updated nodes of shape ([N], F)
        """
        n, eu = inputs
        out, _ = self.gru_cell(eu, n, **kwargs)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(GRUUpdate, self).get_config()
        conf_cell = self.gru_cell.get_config()
        param_list = ["units", "activation", "recurrent_activation",
                      "use_bias", "kernel_initializer",
                      "recurrent_initializer",
                      "bias_initializer", "kernel_regularizer",
                      "recurrent_regularizer", "bias_regularizer", "kernel_constraint",
                      "recurrent_constraint", "bias_constraint", "dropout",
                      "recurrent_dropout", "reset_after"]
        for x in param_list:
            if x in conf_cell.keys():
                config.update({x: conf_cell[x]})
        return config


class ResidualLayer(Layer):
    r"""Residual Layer as defined by `DimNetPP <https://arxiv.org/abs/2011.14115>`__ ."""

    def __init__(self, units,
                 use_bias=True,
                 activation='swish',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        """Initialize layer.

        Args:
            units: Dimension of the kernel.
            use_bias (bool, optional): Use bias. Defaults to True.
            activation (str): Activation function. Default is "kgcnn>swish".
            kernel_regularizer: Kernel regularization. Default is None.
            bias_regularizer: Bias regularization. Default is None.
            activity_regularizer: Activity regularization. Default is None.
            kernel_constraint: Kernel constrains. Default is None.
            bias_constraint: Bias constrains. Default is None.
            kernel_initializer: Initializer for kernels. Default is 'glorot_uniform'.
            bias_initializer: Initializer for bias. Default is 'zeros'.
        """
        super(ResidualLayer, self).__init__(**kwargs)
        dense_args = {
            "units": units, "activation": activation, "use_bias": use_bias,
            "kernel_regularizer": kernel_regularizer, "activity_regularizer": activity_regularizer,
            "bias_regularizer": bias_regularizer, "kernel_constraint": kernel_constraint,
            "bias_constraint": bias_constraint, "kernel_initializer": kernel_initializer,
            "bias_initializer": bias_initializer
        }

        self.dense_1 = Dense(**dense_args)
        self.dense_2 = Dense(**dense_args)
        self.add_end = Add()

    def build(self, input_shape):
        """Build layer."""
        super(ResidualLayer, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (Tensor): Node or edge embedding of shape ([N], F)

        Returns:
            Tensor: Node or edge embedding of shape ([N], F)
        """
        x = self.dense_1(inputs, **kwargs)
        x = self.dense_2(x, **kwargs)
        x = self.add_end([inputs, x], **kwargs)
        return x

    def get_config(self):
        config = super(ResidualLayer, self).get_config()
        conf_dense = self.dense_1.get_config()
        for x in ["kernel_regularizer", "activity_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer", "activation", "use_bias", "units"]:
            if x in conf_dense.keys():
                config.update({x: conf_dense[x]})
        return config



import keras as ks
from kgcnn.layers.message import MessagePassingBase
from kgcnn.layers.norm import GraphBatchNormalization
from keras.layers import Activation, Multiply, Concatenate, Add, Dense


class CGCNNLayer(MessagePassingBase):
    r"""Message Passing Layer used in the Crystal Graph Convolutional Neural Network:

    `CGCNN <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.120.145301>`__ .

    Based on the original code in pytorch (<https://github.com/txie-93/cgcnn>).

    Args:
        units (int): Units for Dense layer.
        activation_s (str): Tensorflow activation applied before gating the message.
        activation_out (str): Tensorflow activation applied the very end of the layer (after gating).
        batch_normalization (bool): Whether to use batch normalization (:obj:`GraphBatchNormalization`) or not.
        use_bias (bool): Boolean, whether the layer uses a bias vector. Default is True.
        kernel_initializer: Initializer for the `kernel` weights matrix. Default is "glorot_uniform".
        bias_initializer: Initializer for the bias vector. Default is "zeros".
        padded_disjoint: Whether disjoint tensors have padded nodes. Default if False.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix. Default is None.
        bias_regularizer: Regularizer function applied to the bias vector. Default is None.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation"). Default is None.
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix. Default is None.
        bias_constraint: Constraint function applied to the bias vector. Default is None.
    """

    def __init__(self, units: int = 64,
                 activation_s="softplus",
                 activation_out="softplus",
                 batch_normalization: bool = False,
                 use_bias: bool = True,
                 kernel_regularizer=None,
                 padded_disjoint: bool = False,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 pooling_method: str = "scatter_mean",
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 **kwargs):
        super(CGCNNLayer, self).__init__(use_id_tensors=4, pooling_method=pooling_method, **kwargs)
        self.units = units
        self.use_bias = use_bias
        self.padded_disjoint = padded_disjoint
        self.batch_normalization = batch_normalization
        kernel_args = {"kernel_regularizer": kernel_regularizer, "bias_regularizer": bias_regularizer,
                       "kernel_constraint": kernel_constraint, "bias_constraint": bias_constraint,
                       "kernel_initializer": kernel_initializer, "bias_initializer": bias_initializer}

        self.activation_f_layer = Activation(activation="sigmoid", activity_regularizer=activity_regularizer)
        self.activation_s_layer = Activation(activation_s, activity_regularizer=activity_regularizer)
        self.activation_out_layer = Activation(activation_out, activity_regularizer=activity_regularizer)
        if batch_normalization:
            self.batch_norm_f = GraphBatchNormalization(padded_disjoint=padded_disjoint)
            self.batch_norm_s = GraphBatchNormalization(padded_disjoint=padded_disjoint)
            self.batch_norm_out = GraphBatchNormalization(padded_disjoint=padded_disjoint)
        self.f = Dense(self.units, activation="linear", use_bias=use_bias, **kernel_args)
        self.s = Dense(self.units, activation="linear", use_bias=use_bias, **kernel_args)
        self.lazy_mult = Multiply()
        self.lazy_add = Add()
        self.lazy_concat = Concatenate(axis=-1)

    def message_function(self, inputs, **kwargs):
        r"""Prepare messages.

        Args:
            inputs: [nodes_in, nodes_out, edges, graph_id_node, graph_id_edge, count_nodes, count_edges]

                - nodes_in (Tensor): Embedding of sending nodes of shape `([M], F)`
                - nodes_out (Tensor): Embedding of sending nodes of shape `([M], F)`
                - edges (Tensor): Embedding of edges of shape `([M], E)`
                - graph_id_node (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - graph_id_edge (Tensor): ID tensor of batch assignment in disjoint graph of shape `([M], )` .
                - nodes_count (Tensor): Tensor of number of nodes for each graph of shape `(batch, )` .
                - edges_count (Tensor): Tensor of number of edges for each graph of shape `(batch, )` .

        Returns:
            Tensor: Messages for updates of shape `([M], units)`.
        """
        nodes_in = inputs[0]  # shape: (batch_size, M, F)
        nodes_out = inputs[1]  # shape: (batch_size, M, F)
        edge_features = inputs[2]  # shape: (batch_size, M, E)
        graph_id_node, graph_id_edge, count_nodes, count_edges = inputs[3:]

        x = self.lazy_concat([nodes_in, nodes_out, edge_features], **kwargs)
        x_s, x_f = self.s(x, **kwargs), self.f(x, **kwargs)
        if self.batch_normalization:
            x_s = self.batch_norm_s([x_s, graph_id_edge, count_edges], **kwargs)
            x_f = self.batch_norm_f([x_f, graph_id_edge, count_edges], **kwargs)
        x_s, x_f = self.activation_s_layer(x_s, **kwargs), self.activation_f_layer(x_f, **kwargs)
        x_out = self.lazy_mult([x_s, x_f], **kwargs)  # shape: (batch_size, M, self.units)
        return x_out

    def update_nodes(self, inputs, **kwargs):
        """Update node embeddings.

        Args:
            inputs: [nodes, nodes_updates, graph_id_node, graph_id_edge, count_nodes, count_edges]

                - nodes (Tensor): Embedding of nodes of previous layer of shape `([M], F)`
                - nodes_updates (Tensor): Node updates of shape `([M], F)`
                - graph_id_node (Tensor): ID tensor of batch assignment in disjoint graph of shape `([N], )` .
                - graph_id_edge (Tensor): ID tensor of batch assignment in disjoint graph of shape `([M], )` .
                - nodes_count (Tensor): Tensor of number of nodes for each graph of shape `(batch, )` .
                - edges_count (Tensor): Tensor of number of edges for each graph of shape `(batch, )` .

        Returns:
            Tensor: Updated nodes of shape `([N], F)`.
        """
        nodes = inputs[0]
        nodes_update = inputs[1]
        graph_id_node, graph_id_edge, count_nodes, count_edges = inputs[2:]

        if self.batch_normalization:
            nodes_update = self.batch_norm_out([nodes_update, graph_id_node, count_nodes], **kwargs)

        nodes_updated = self.lazy_add([nodes, nodes_update], **kwargs)
        nodes_updated = self.activation_out_layer(nodes_updated, **kwargs)
        return nodes_updated

    def get_config(self):
        """Update layer config."""
        config = super(CGCNNLayer, self).get_config()
        config.update({
            "units": self.units, "use_bias": self.use_bias, "padded_disjoint": self.padded_disjoint,
            "batch_normalization": self.batch_normalization})
        conf_s = self.activation_s_layer.get_config()
        conf_out = self.activation_out_layer.get_config()
        config.update({"activation_s": conf_s["activation"]})
        config.update({"activation_out": conf_out["activation"]})
        if "activity_regularizer" in conf_out.keys():
            config.update({"activity_regularizer": conf_out["activity_regularizer"]})
        conf_f = self.f.get_config()
        for x in ["kernel_regularizer", "bias_regularizer", "kernel_constraint",
                  "bias_constraint", "kernel_initializer", "bias_initializer"]:
            if x in conf_f.keys():
                config.update({x: conf_f[x]})
        return config



import keras as ks
from keras import ops
from kgcnn.ops.scatter import scatter_reduce_sum, scatter_reduce_max
from kgcnn.layers.aggr import Aggregate

# Order Matters: Sequence to sequence for sets
# by Vinyals et al. 2016
# https://arxiv.org/abs/1511.06391


class PoolingSet2SetEncoder(ks.layers.Layer):
    r"""Pooling Node or edge embeddings by the Set2Set encoder part from layer.
    This was first proposed by `NMPNN <http://arxiv.org/abs/1704.01212>`__ .
    The Reading to Memory has to be handled separately.
    Uses a keras LSTM layer for the updates.
    """

    def __init__(self,
                 # Args
                 channels,
                 T=3,  # noqa
                 pooling_method='mean',
                 init_qstar='mean',
                 # Args for LSTM
                 activation="tanh",
                 recurrent_activation="sigmoid",
                 use_bias=True,
                 kernel_initializer="glorot_uniform",
                 recurrent_initializer="orthogonal",
                 bias_initializer="zeros",
                 unit_forget_bias=True,
                 kernel_regularizer=None,
                 recurrent_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 recurrent_constraint=None,
                 bias_constraint=None,
                 dropout=0.0,
                 recurrent_dropout=0.0,
                 implementation=2,
                 return_sequences=False,  # Should not be changed here
                 return_state=False,  # Should not be changed here
                 go_backwards=False,  # Should not be changed here
                 stateful=False,
                 # time_major=False,
                 unroll=False,
                 **kwargs):
        """Initialize layer.

        Args:
            channels (int): Number of channels for the LSTM update.
            T (int): Numer of iterations. Default is T=3.
            pooling_method : Pooling method for PoolingSet2SetEncoder. Default is 'mean'.
            init_qstar: How to generate the first q_star vector. Default is 'mean'.
            activation: Activation function to use.
                Default: hyperbolic tangent (`tanh`). If you pass `None`, no activation
                is applied (ie. "linear" activation: `a(x) = x`).
            recurrent_activation: Activation function to use for the recurrent step.
                Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
                applied (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean (default `True`), whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix, used for
                the linear transformation of the inputs. Default: `glorot_uniform`.
                recurrent_initializer: Initializer for the `recurrent_kernel` weights
                matrix, used for the linear transformation of the recurrent state.
                Default: `orthogonal`.
            bias_initializer: Initializer for the bias vector. Default: `zeros`.
                unit_forget_bias: Boolean (default `True`). If True, add 1 to the bias of
                the forget gate at initialization. Setting it to true will also force
                `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
                al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf).
            kernel_regularizer: Regularizer function applied to the `kernel` weights
                matrix. Default: `None`.
            recurrent_regularizer: Regularizer function applied to the
                `recurrent_kernel` weights matrix. Default: `None`.
                bias_regularizer: Regularizer function applied to the bias vector. Default:
                `None`.
            activity_regularizer: Regularizer function applied to the output of the
                layer (its "activation"). Default: `None`.
            kernel_constraint: Constraint function applied to the `kernel` weights
                matrix. Default: `None`.
            recurrent_constraint: Constraint function applied to the `recurrent_kernel`
                weights matrix. Default: `None`.
            bias_constraint: Constraint function applied to the bias vector. Default:
                `None`.
            dropout: Float between 0 and 1. Fraction of the units to drop for the linear
                transformation of the inputs. Default: 0.
                recurrent_dropout: Float between 0 and 1. Fraction of the units to drop for
                the linear transformation of the recurrent state. Default: 0.
            return_sequences: Boolean. Whether to return the last output. in the output
                sequence, or the full sequence. Default: `False`.
            return_state: Boolean. Whether to return the last state in addition to the
                output. Default: `False`.
            go_backwards: Boolean (default `False`). If True, process the input sequence
                backwards and return the reversed sequence.
            stateful: Boolean (default `False`). If True, the last state for each sample
                at index i in a batch will be used as initial state for the sample of
                index i in the following batch.
            unroll: Boolean (default `False`). If True, the network will be unrolled,
                else a symbolic loop will be used. Unrolling can speed-up a RNN, although
                it tends to be more memory-intensive. Unrolling is only suitable for short
                sequences.
        """
        super(PoolingSet2SetEncoder, self).__init__(**kwargs)
        # Number of Channels to use in LSTM
        self.channels = channels
        self.T = T  # Number of Iterations to work on memory
        self.pooling_method = pooling_method
        self.init_qstar = init_qstar

        # Reduction of messages for f_et
        self._reduce_keys = {
            "sum": ops.sum,
            "mean": ops.mean,
            "max": ops.max,
            "min": ops.min,
            "var": ops.var,
        }
        if self.pooling_method not in self._reduce_keys:
            raise ValueError("ERROR:kgcnn: Unknown reduction '%s', choose one of '%s'." % (
                self.pooling_method, self._reduce_keys.keys()))
        self._reduce = self._reduce_keys[self.pooling_method]

        self._pool_init = None
        if self.init_qstar in ["0", "zeros", "zero"]:
            self.qstar0 = self.init_qstar_0
        elif self.init_qstar in ["ref", "reference", "input"]:
            self.qstar0 = self.init_qstar_ref
        else:
            self._pool_init = Aggregate(pooling_method=self.init_qstar)
            self.qstar0 = self.init_qstar_pool

        # LSTM Layer to run on m
        self.lay_lstm = ks.layers.LSTM(
            channels,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            unit_forget_bias=unit_forget_bias,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            implementation=implementation,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            # time_major=time_major,
            unroll=unroll
        )

    def build(self, input_shape):
        """Build layer."""
        assert len(input_shape) == 3
        ref_shape, attr_shape, index_shape = [list(x) for x in input_shape]
        if self._pool_init is not None:
            self._pool_init.build([attr_shape, index_shape, ref_shape])
        self.lay_lstm.build(tuple(ref_shape[:1] + [1] + [2*self.channels]))
        self.built = True

    def compute_output_shape(self, input_shape):
        assert len(input_shape) == 3
        ref_shape, attr_shape, index_shape = [list(x) for x in input_shape]
        return tuple(ref_shape[:1] + [1] + [2*self.channels])

    def call(self, inputs, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [reference, nodes, batch_index]

                - reference (Tensor): Reference for aggregation of shape `(batch, ...)` .
                - nodes (Tensor): Node embeddings of shape `([N], F)` .
                - batch_index (Tensor): Batch assignment of shape `([N], )` .

        Returns:
            Tensor: Pooled tensor q_star of shape `(batch, 1, 2*channels)`
        """
        ref, x, batch_index = inputs

        # Reading to memory removed here, is to be done by separately
        m = x  # ([N], feat)

        # Initialize q0 and r0
        q_star = self.qstar0(m, batch_index, ref)

        # start loop
        for i in range(0, self.T):
            q = self.lay_lstm(q_star)  # (batch, feat)
            q_star = self.update_q(q, m, batch_index, ref)

        return q_star

    def update_q(self, q, m, batch_index, ref):
        qt = ops.take(q, batch_index, axis=0)
        et = self.f_et(m, qt)  # (batch*num,)
        # get at = exp(et)/sum(et) with sum(et)
        at = ops.exp(et - self._get_scale_per_sample(et, batch_index, ref))  # (batch*num,)
        norm = self._get_norm(at, batch_index, ref)  # (batch*num,)
        at = norm * at  # (batch*num,) x (batch*num,)
        # calculate rt
        # at = ops.expand_dims(at, axis=1)
        rt = m * at  # (batch*num,feat) x (batch*num,1)
        rt = self._pool_sum([rt, batch_index, ref])  # (batch,feat)
        # qstar = [q,r]
        q_star = ops.concatenate([q, rt], axis=1)  # (batch,2*feat)
        q_star = ops.expand_dims(q_star, axis=1)  # (batch,1,2*feat)
        return q_star

    def f_et(self, fm, fq):
        r"""Function to compute scalar from :math:`m` and :math:`q` .
        Uses :obj:`pooling_method` argument of the layer.

        Args:
             fm (Tensor): of shape `([N], F)` .
             fq (Tensor): of shape `([N], F)` .

        Returns:
            Tensor: et of shape `([N], )` .
        """
        return self._reduce(fm * fq, axis=1, keepdims=True)  # ([N], )

    @staticmethod
    def _get_scale_per_batch(x):
        """Get re-scaling for the batch."""
        return ops.max(x, axis=0, keepdims=True)

    def _get_scale_per_sample(self, x, ind, ref):
        """Get re-scaling for the sample."""
        out = self._pool_max([x, ind, ref])  # (batch,)
        out = ops.take(out, ind, axis=0)  # (batch*num,)
        return out

    def _get_norm(self, x, ind, ref):
        """Compute Norm."""
        norm = self._pool_sum([x, ind, ref])  # (batch,)
        norm = ops.divide(ops.convert_to_tensor(1.0, dtype=norm.dtype), norm)  # (batch,)
        norm = ops.where(ops.logical_or(ops.isnan(norm), ops.isinf(norm)), 0., norm)
        norm = ops.take(norm, ind, axis=0)  # (batch*num,)
        return norm

    def init_qstar_ref(self, m, batch_index, reference):
        return reference

    def init_qstar_0(self, m, batch_index, reference):
        """Initialize the q0 with zeros."""
        batch_shape = ops.shape(reference)[0]
        return ops.zeros((batch_shape, 1, 2 * self.channels), dtype=m.dtype)

    def init_qstar_pool(self, m, batch_index, reference):
        """Initialize the q0 with mean."""
        # batch_shape = ksb.shape(batch_num)
        q = self._pool_init([m, batch_index, reference])  # (batch,feat)
        qstar= self.update_q(q, m, batch_index, reference)
        return qstar

    def get_config(self):
        """Make config for layer."""
        config = super(PoolingSet2SetEncoder, self).get_config()
        config.update({"channels": self.channels, "T": self.T, "pooling_method": self.pooling_method,
                       "init_qstar": self.init_qstar})
        lstm_conf = self.lay_lstm.get_config()
        lstm_param = ["activation",
                      "recurrent_activation",
                      "use_bias",
                      "kernel_initializer",
                      "recurrent_initializer",
                      "bias_initializer",
                      "unit_forget_bias",
                      "kernel_regularizer",
                      "recurrent_regularizer",
                      "bias_regularizer",
                      "activity_regularizer",
                      "kernel_constraint",
                      "recurrent_constraint",
                      "bias_constraint",
                      "dropout",
                      "recurrent_dropout",
                      "implementation",
                      "return_sequences",  # Should not be changed here
                      "return_state",  # Should not be changed here
                      "go_backwards",  # Should not be changed here
                      "stateful",
                      # "time_major",
                      "unroll"]
        for x in lstm_param:
            if x in lstm_conf.keys():
                config.update({x: lstm_conf[x]})
        return config

    def _pool_sum(self, inputs):
        values, indices, ref = inputs
        shape_ = ops.shape(ref)[:1] + ops.shape(values)[1:]
        return scatter_reduce_sum(indices, values, shape=shape_)

    def _pool_max(self, inputs):
        values, indices, ref = inputs
        shape_ = ops.shape(ref)[:1] + ops.shape(values)[1:]
        return scatter_reduce_max(indices, values, shape=shape_)


import keras as ks
from keras import ops


class Embedding(ks.layers.Layer):

    def __init__(self,
                 input_dim,
                 output_dim,
                 embeddings_initializer='uniform',
                 embeddings_regularizer=None,
                 activity_regularizer=None,
                 embeddings_constraint=None,
                 mask_zero=False,
                 input_length=None,
                 sparse=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim

        self.embeddings_initializer = ks.initializers.get(embeddings_initializer)
        self.embeddings_regularizer = ks.regularizers.get(embeddings_regularizer)
        self.activity_regularizer = ks.regularizers.get(activity_regularizer)
        self.embeddings_constraint = ks.constraints.get(embeddings_constraint)

        self.embeddings = self.add_weight(
            name="embeddings",
            shape=(input_dim, output_dim),
            dtype=self.dtype,
            initializer=self.embeddings_initializer,
            regularizer=self.embeddings_regularizer,
            constraint=self.embeddings_constraint,
            trainable=True,
        )

    def build(self, input_shape):
        super(Embedding, self).build(input_shape)

    def call(self, inputs):
        return ops.take(self.embeddings, inputs, axis=0)

    def get_config(self):
        return super(Embedding, self).get_config()


class ExpandDims(ks.layers.Layer):

    def __init__(self, axis, **kwargs):
        super(ExpandDims, self).__init__(**kwargs)
        self.axis = axis

    def build(self, input_shape):
        self.built = True

    def call(self, inputs):
        return ops.expand_dims(inputs, axis=self.axis)

    def get_config(self):
        config = super(ExpandDims, self).get_config()
        config.update({"axis": self.axis})
        return config


class SqueezeDims(ks.layers.Layer):

    def __init__(self, axis, **kwargs):
        super(SqueezeDims, self).__init__(**kwargs)
        self.axis = axis

    def build(self, input_shape):
        self.built = True

    def call(self, inputs):
        return ops.squeeze(inputs, axis=self.axis)

    def get_config(self):
        config = super(SqueezeDims, self).get_config()
        config.update({"axis": self.axis})
        return config


def Input(
        shape=None,
        batch_size=None,
        dtype=None,
        sparse=None,
        batch_shape=None,
        name=None,
        tensor=None,
        ragged=None
    ):

    layer = ks.layers.InputLayer(
        shape=shape,
        batch_size=batch_size,
        dtype=dtype,
        sparse=sparse,
        batch_shape=batch_shape,
        name=name,
        input_tensor=tensor,
    )
    return layer.output


class ZerosLike(ks.layers.Layer):
    r"""Layer to make a zero tensor"""

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(ZerosLike, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build layer."""
        super(ZerosLike, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (Tensor): Tensor of node or edge embeddings of shape ([N], F, ...)

        Returns:
            Tensor: Zero-like tensor of input.
        """
        return ops.zeros_like(inputs)


from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.modules import Embedding
from keras.layers import Dense, Dropout, Add
from kgcnn.layers.pooling import PoolingNodes
from ._layers import rGIN


def model_disjoint(
    inputs,
    use_node_embedding,
    input_node_embedding,
    gin_mlp,
    depth,
    rgin_args,
    last_mlp,
    output_mlp,
    output_embedding,
    dropout
):
    n, edi, batch_id_node, count_nodes = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)

    # Model
    # Map to the required number of units.
    n_units = gin_mlp["units"][-1] if isinstance(gin_mlp["units"], list) else int(gin_mlp["units"])
    n = Dense(n_units, use_bias=True, activation='linear')(n)
    list_embeddings = [n]
    for i in range(0, depth):
        n = rGIN(**rgin_args)([n, edi])
        n = GraphMLP(**gin_mlp)([n, batch_id_node, count_nodes])
        list_embeddings.append(n)

    # Output embedding choice
    if output_embedding == "graph":
        out = [PoolingNodes()([count_nodes, x, batch_id_node]) for x in list_embeddings]  # will return tensor
        out = [MLP(**last_mlp)(x) for x in out]
        out = [Dropout(dropout)(x) for x in out]
        out = Add()(out)
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":  # Node labeling
        out = n
        out = GraphMLP(**last_mlp)([out, batch_id_node, count_nodes])
        out = GraphMLP(**output_mlp)([out, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `rGIN`")

    return out



from keras.layers import Concatenate
from kgcnn.layers.aggr import AggregateLocalEdgesLSTM, AggregateLocalEdges
from kgcnn.layers.gather import GatherNodesOutgoing
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.norm import GraphLayerNormalization
from kgcnn.layers.pooling import PoolingNodes


def model_disjoint(
        inputs,
        use_node_embedding: bool = None,
        use_edge_embedding: bool = None,
        input_node_embedding: dict = None,
        input_edge_embedding: dict = None,
        node_mlp_args: dict = None,
        edge_mlp_args: dict = None,
        pooling_args: dict = None,
        pooling_nodes_args: dict = None,
        gather_args: dict = None,
        concat_args: dict = None,
        use_edge_features: bool = None,
        depth: int = None,
        output_embedding: str = None,
        output_mlp: dict = None,
):
    n, ed, disjoint_indices, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    # Embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)

    for i in range(0, depth):

        eu = GatherNodesOutgoing(**gather_args)([n, disjoint_indices])
        if use_edge_features:
            eu = Concatenate(**concat_args)([eu, ed])

        eu = GraphMLP(**edge_mlp_args)([eu, batch_id_edge, count_edges])

        # Pool message
        if pooling_args['pooling_method'] in ["LSTM", "lstm"]:
            nu = AggregateLocalEdgesLSTM(**pooling_args)([n, eu, disjoint_indices])
        else:
            nu = AggregateLocalEdges(**pooling_args)([n, eu, disjoint_indices])  # Summing for each node connection

        nu = Concatenate(**concat_args)([n, nu])  # Concatenate node features with new edge updates

        n = GraphMLP(**node_mlp_args)([nu, batch_id_node, count_nodes])

        n = GraphLayerNormalization()([n, batch_id_node, count_nodes])

    # Regression layer on output
    if output_embedding == 'graph':
        out = PoolingNodes(**pooling_nodes_args)([count_nodes, n, batch_id_node])
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for `GraphSAGE`")
    return out


from keras import ops
from kgcnn.layers.gather import GatherNodesOutgoing, GatherEdgesPairs
from kgcnn.layers.aggr import AggregateLocalEdges
from keras.layers import Subtract, Add, Layer


class DMPNNPPoolingEdgesDirected(Layer):  # noqa
    """Pooling of edges for around a target node as defined by

    `DMPNN <https://pubs.acs.org/doi/full/10.1021/acs.jcim.9b00237>`__ . This is slightly different as the normal node
    aggregation from message passing like networks. Requires edge pair indices for this implementation.
    """

    def __init__(self, **kwargs):
        """Initialize layer."""
        super(DMPNNPPoolingEdgesDirected, self).__init__(**kwargs)
        self.pool_edge_1 = AggregateLocalEdges(pooling_method="scatter_sum")
        self.gather_edges = GatherNodesOutgoing()
        self.gather_pairs = GatherEdgesPairs()
        self.subtract_layer = Subtract()

    def build(self, input_shape):
        super(DMPNNPPoolingEdgesDirected, self).build(input_shape)
        # Could call build on sub-layers but is not necessary.

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs: [nodes, edges, edge_index, edge_reverse_pair]

                - nodes (Tensor): Node embeddings of shape ([N], F)
                - edges (Tensor): Edge or message embeddings of shape ([M], F)
                - edge_index (Tensor): Edge indices referring to nodes of shape (2, [M])
                - edge_reverse_pair (Tensor): Pair mappings for reverse edges (1, [M])

        Returns:
            Tensor: Edge embeddings of shape ([M], F)
        """
        n, ed, edi, edp = inputs
        pool_edge_receive = self.pool_edge_1([n, ed, edi], **kwargs)  # Sum pooling of all edges
        ed_new = self.gather_edges([pool_edge_receive, edi], **kwargs)
        ed_not = self.gather_pairs([ed, edp],  **kwargs)
        out = self.subtract_layer([ed_new, ed_not],  **kwargs)
        return out


class GIN_D(Layer):
    r"""Convolutional unit of `Graph Isomorphism Network from: How Powerful are Graph Neural Networks?
    <https://arxiv.org/abs/1810.00826>`__ .

    Modified to use :math:`h_{w_0}`

    Computes graph convolution at step :math:`k` for node embeddings :math:`h_\nu` as:

    .. math::

        h_\nu^{(k)} = \phi^{(k)} ((1+\epsilon^{(k)}) h_\nu^{0} + \sum_{u\in N(\nu)}) h_u^{k-1}.

    with optional learnable :math:`\epsilon^{(k)}`

    .. note::

        The non-linear mapping :math:`\phi^{(k)}`, usually an :obj:`MLP`, is not included in this layer.
    """

    def __init__(self,
                 pooling_method='sum',
                 epsilon_learnable=False,
                 **kwargs):
        """Initialize layer.

        Args:
            epsilon_learnable (bool): If epsilon is learnable or just constant zero. Default is False.
            pooling_method (str): Pooling method for summing edges. Default is 'segment_sum'.
        """
        super(GIN_D, self).__init__(**kwargs)
        self.pooling_method = pooling_method
        self.epsilon_learnable = epsilon_learnable

        # Layers
        self.lay_gather = GatherNodesOutgoing()
        self.lay_pool = AggregateLocalEdges(pooling_method=self.pooling_method)
        self.lay_add = Add()

        # Epsilon with trainable as optional and default zeros initialized.
        self.eps_k = self.add_weight(name="epsilon_k", trainable=self.epsilon_learnable,
                                     initializer="zeros", dtype=self.dtype)

    def build(self, input_shape):
        """Build layer."""
        super(GIN_D, self).build(input_shape)

    def call(self, inputs, **kwargs):
        r"""Forward pass.
        Args:
            inputs: [nodes, edge_index, nodes_0]

                - nodes (Tensor): Node embeddings of shape `([N], F)`
                - edge_index (Tensor): Edge indices referring to nodes of shape `(2, [M])`
                - nodes_0 (Tensor): Node embeddings of shape `([N], F)`

        Returns:
            Tensor: Node embeddings of shape `([N], F)`
        """
        # Need to check if edge_index is full and not half (directed).
        node, edge_index, node_0 = inputs
        ed = self.lay_gather([node, edge_index], **kwargs)
        # Summing for each node connection
        nu = self.lay_pool([node, ed, edge_index], **kwargs)
        # Modified to use node_0 instead of node see equation 7 in paper.
        no = (ops.convert_to_tensor(1, dtype=self.eps_k.dtype) + self.eps_k) * node_0
        out = self.lay_add([no, nu], **kwargs)
        return out

    def get_config(self):
        """Update config."""
        config = super(GIN_D, self).get_config()
        config.update({"pooling_method": self.pooling_method,
                       "epsilon_learnable": self.epsilon_learnable})
        return config



import torch


def scatter_reduce_sum(indices, values, shape):
    dims_to_add = values.dim() - indices.dim()
    for _ in range(dims_to_add):
        indices = torch.unsqueeze(indices, dim=-1)
    return torch.zeros(*shape, dtype=values.dtype, device=values.device).scatter_reduce(
        0, torch.broadcast_to(indices, values.shape), values, reduce='sum')


def scatter_reduce_min(indices, values, shape):
    dims_to_add = values.dim() - indices.dim()
    for _ in range(dims_to_add):
        indices = torch.unsqueeze(indices, dim=-1)
    return torch.zeros(*shape, dtype=values.dtype, device=values.device).scatter_reduce(
        0, torch.broadcast_to(indices, values.shape), values, reduce='amin', include_self=False)


def scatter_reduce_max(indices, values, shape):
    dims_to_add = values.dim() - indices.dim()
    for _ in range(dims_to_add):
        indices = torch.unsqueeze(indices, dim=-1)
    return torch.zeros(*shape, dtype=values.dtype, device=values.device).scatter_reduce(
        0, torch.broadcast_to(indices, values.shape), values, reduce='amax', include_self=False)


def scatter_reduce_mean(indices, values, shape):
    dims_to_add = values.dim() - indices.dim()
    for _ in range(dims_to_add):
        indices = torch.unsqueeze(indices, dim=-1)
    return torch.zeros(*shape, dtype=values.dtype, device=values.device).scatter_reduce(
        0, torch.broadcast_to(indices, values.shape), values, reduce='mean', include_self=False)


def scatter_reduce_softmax(indices, values, shape, normalize: bool = False):
    indices_scatter = indices
    dims_to_add = values.dim()-indices.dim()
    for _ in range(dims_to_add):
        indices_scatter = torch.unsqueeze(indices_scatter, dim=-1)

    if normalize:
        zeros_min = torch.zeros(*shape, dtype=values.dtype, device=values.device)
        data_segment_max = zeros_min.scatter_reduce(
            0, torch.broadcast_to(indices_scatter, values.shape), values, reduce='amax', include_self=False)
        data_max = torch.index_select(data_segment_max, dim=0, index=indices)
        values = values - data_max

    values_exp = torch.exp(values)
    zeros = torch.zeros(*shape, dtype=values.dtype, device=values.device)
    values_exp_sum = zeros.scatter_reduce(
        0, torch.broadcast_to(indices_scatter, values_exp.shape), values_exp, reduce='sum', include_self=True)
    values_exp_sum = torch.index_select(values_exp_sum, dim=0, index=indices)
    return values_exp / values_exp_sum


def repeat_static_length(x, repeats, axis=None, total_repeat_length: int = None):
    # from keras_core.backend.torch.numpy import repeat
    return torch.repeat_interleave(x, repeats, dim=axis)


def decompose_ragged_tensor(x):
    raise NotImplementedError("Operation supported this backend '%s'." % __name__)


def norm(x, ord='fro', axis=None, keepdims=False):
    return torch.linalg.norm(x, ord=ord, dim=axis, keepdims=keepdims)


def cross(x1, x2):
    return torch.cross(x1, x2, dim=-1)



import keras as ks
import kgcnn.ops.activ
from kgcnn.layers.geom import (
    DisplacementVectorsUnitCell,
    DisplacementVectorsASU, NodePosition, FracToRealCoordinates,
    EuclideanNorm, GaussBasisLayer
)
from kgcnn.layers.mlp import MLP
from kgcnn.layers.modules import Embedding
from kgcnn.layers.pooling import PoolingWeightedNodes, PoolingNodes
from ._layers import CGCNNLayer


def model_disjoint_crystal(
        inputs: list,
        representation=None,
        use_node_embedding=None,
        output_embedding=None,
        input_node_embedding=None,
        expand_distance=None,
        conv_layer_args=None,
        make_distances=None,
        depth=None,
        gauss_args=None,
        node_pooling_args=None,
        output_mlp=None
):
    atom_attributes, multiplicities, coord, symmops, edge_indices, cell_translations, lattice_matrix, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    if make_distances:

        frac_coords = coord

        if representation == 'unit':
            displacement_vectors = DisplacementVectorsUnitCell()([frac_coords, edge_indices, cell_translations])

        elif representation == 'asu':
            displacement_vectors = DisplacementVectorsASU()([frac_coords, edge_indices, symmops, cell_translations])
        else:
            x_in, x_out = NodePosition()([frac_coords, edge_indices])
            displacement_vectors = ks.layers.Subtract()([x_out, x_in])

        displacement_vectors = FracToRealCoordinates()([displacement_vectors, lattice_matrix, batch_id_edge])

        edge_distances = EuclideanNorm(axis=-1, keepdims=True)(displacement_vectors)

    else:
        edge_distances = coord

    if expand_distance:
        edge_distances = GaussBasisLayer(**gauss_args)(edge_distances)

    # embedding, if no feature dimension
    if use_node_embedding:
        n = Embedding(**input_node_embedding)(atom_attributes)
    else:
        n = atom_attributes

    n = ks.layers.Dense(conv_layer_args["units"], activation='linear')(n)
    for _ in range(depth):
        n = CGCNNLayer(**conv_layer_args)([
            n, edge_distances, edge_indices, batch_id_node, batch_id_edge, count_nodes, count_edges
        ])

    if representation == 'asu':
        out = PoolingWeightedNodes(**node_pooling_args)([count_nodes, n, multiplicities, batch_id_node])
    else:
        out = PoolingNodes(**node_pooling_args)([count_nodes, n, batch_id_node])

    out = MLP(**output_mlp)(out)

    # Only graph embedding for CGCNN.
    if output_embedding != "graph":
        raise ValueError("Unsupported output embedding for mode `CGCNN` .")

    return out



import logging
import keras as ks
from kgcnn.models.utils import get_model_class


def deserialize(obj_dict: dict) -> ks.models.Model:

    if isinstance(obj_dict, ks.models.Model):
        return obj_dict
    if not isinstance(obj_dict, dict):
        raise ValueError("Can not deserialize model with '%s'." % obj_dict)
    if "config" not in obj_dict:
        obj_dict.update({"config": {}})
    class_name = obj_dict.get("class_name")
    module_name = obj_dict.get("module_name")
    model_cls = get_model_class(module_name=module_name, class_name=class_name)
    obj = model_cls(**obj_dict["config"])

    if hasattr(obj, "set_weights") and "weights" in obj_dict:
        obj.set_weights(obj_dict["weights"])

    # Call class methods if methods are in obj_dict.
    # Order is important here.
    if "methods" in obj_dict:
        method_list = obj_dict["methods"]
        if hasattr(obj, "set_methods"):
            obj.set_methods(method_list)
        else:
            # Try setting them manually.
            for method_item in method_list:
                for method, kwargs in method_item.items():
                    if hasattr(obj, method):
                        getattr(obj, method)(**kwargs)
                    else:
                        logging.error("Class for deserialization does not have method '%s'." % method)

    return obj



import keras as ks
import logging
from typing import Union
import numpy as np
from numpy.random import Generator, PCG64
import tensorflow as tf


# Module logger
logging.basicConfig()
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


def pad_at_axis(x, pad_width, axis=0, **kwargs):
    pads = [(0, 0) for _ in range(len(x.shape))]
    pads[axis] = pad_width
    return np.pad(x, pad_width=pads, **kwargs)


def tf_dataset_disjoint_generator(
        graphs,
        inputs: Union[list, dict],
        assignment_to_id: Union[list, dict] = None,
        assignment_of_indices: Union[list, dict] = None,
        pos_batch_id: Union[list, dict] = None,
        pos_subgraph_id: Union[list, dict] = None,
        pos_count: Union[list, dict] = None,
        batch_size=32,
        epochs=None,
        padded_disjoint=False,
        shuffle=True,
        seed=42
):
    r"""Make a tensorflow dataset for disjoint graph loading.

    For the moment only IDs that have their values in inputs can be generated, as the value tensors of e.g. node
    or edge are used to generate batch IDs.

    Inputs is a list or dictionary of keras input layer configs. The names of the layers are linked to the properties
    in `graph` .

    With `assignment_to_id` and `assignment_of_indices` disjoint indices and attributes can be defined.
    Their IDs are marked with `pos_batch_id` etc. One must use a name or index for each general split, since for
    example edge IDs can be used for edge indices, edge attributes and edge relation tensors at the same time.
    Therefore, one batch ID for edges is enough. One could however assign as many as IDs as there are disjoint
    graph properties in `graph` .

    Args:
        graphs: List of dictionaries with named graph properties.
        inputs: List or dict of keras input layer configs.
        assignment_to_id: Assignment of if inputs to disjoint properties to IDs.
        assignment_of_indices: Assignment of inputs (if they are indices) to their reference.
        pos_batch_id: Position or name of batch IDs.
        pos_subgraph_id: Position or name of batch IDs.
        pos_count: Position or name of batch IDs.
        batch_size: Batch size.
        epochs: Expected number of epochs. Only required for padded disjoint.
        padded_disjoint: If padded disjoint tensors should be generated.
        shuffle: Whether to shuffle each epoch.
        seed: Seed for shuffle.

    Returns:
        tf.data.Dataset: Tensorflow dataset to load disjoint graphs.
    """
    # Stats on the required dataset.
    dataset_size = len(graphs)
    data_index = np.arange(dataset_size)
    num_inputs = len(inputs)

    # Check input information for outputspec.
    is_single_input = False
    is_list_input = False
    if isinstance(inputs, list):
        is_list_input = True
        output_spec = tuple([tf.TensorSpec(shape=tuple([None] + list(x["shape"])), dtype=x["dtype"]) for x in inputs])
    elif isinstance(inputs, dict):
        if "shape" in inputs and "dtype" in inputs:
            output_spec = tf.TensorSpec(shape=tuple([None] + list(inputs["shape"])), dtype=inputs["dtype"])
            inputs = {0: inputs}
            is_single_input = True
            num_inputs = 1
        else:
            output_spec = dict(
                {i: tf.TensorSpec(shape=tuple([None] + list(x["shape"])), dtype=x["dtype"]) for i, x in inputs.items()})
    else:
        raise ValueError("Inputs must be list or dict of keras input layer kwargs.")

    # We use a dict for both list and dict input.
    def _convert_to_dict(container_to_check):
        if container_to_check is None:
            return {}
        if isinstance(container_to_check, (list, tuple)):
            return {i: x for i, x in enumerate(container_to_check)}
        if not isinstance(container_to_check, dict):
            raise ValueError("Must be dict or list for mapping and containers.")
        return container_to_check

    inputs = _convert_to_dict(inputs)
    assignment_to_id = _convert_to_dict(assignment_to_id)
    assignment_of_indices = _convert_to_dict(assignment_of_indices)
    pos_batch_id = _convert_to_dict(pos_batch_id)
    pos_subgraph_id = _convert_to_dict(pos_subgraph_id)
    pos_count = _convert_to_dict(pos_count)

    # Fill assignments with Nones if they are not used for input.
    if len(assignment_to_id) < num_inputs:
        for key, values in inputs.items():
            if key not in assignment_to_id.keys():
                assignment_to_id[key] = None
    if len(assignment_of_indices) < num_inputs:
        for key, values in inputs.items():
            if key not in assignment_of_indices.keys():
                assignment_of_indices[key] = None

    flag_batch_id = {i: None for i in inputs.keys()}
    for i, x in pos_batch_id.items():
        flag_batch_id[x] = i

    flag_count = {i: None for i in inputs.keys()}
    for i, x in pos_count.items():
        flag_count[x] = i

    flag_subgraph_id = {i: None for i in inputs.keys()}
    for i, x in pos_subgraph_id.items():
        flag_subgraph_id[x] = i

    all_flags = [flag_batch_id, flag_count, flag_subgraph_id]
    is_attributes = {i: True if all([x[i] is None for x in all_flags]) else False for i in inputs.keys()}

    max_size = {i: [] if assignment_to_id[i] is not None else None for i in inputs.keys()}
    total_max = {i: [] if assignment_to_id[i] is not None else None for i in inputs.keys()}

    # We can check the maximum batch size at the beginning or just have a maximum batch size for each epoch.
    if padded_disjoint:
        if epochs is None:
            raise ValueError("Requires number of epochs if `padded_disjoint=True` .")

        for i in inputs.keys():
            if assignment_to_id[i] is None:
                continue
            len_list = [len(x[inputs[i]["name"]]) for x in graphs]
            total_max[i] = max(len_list)

        rng = Generator(PCG64(seed=seed))

        for epoch in range(epochs):
            max_size_epoch = {i: [] if assignment_to_id[i] is not None else None for i in inputs.keys()}
            if shuffle:
                rng.shuffle(data_index)
            for batch_index in range(0, dataset_size, batch_size):
                idx = data_index[batch_index:batch_index + batch_size]
                graphs_batch = [graphs[i] for i in idx]
                for i in inputs.keys():
                    if assignment_to_id[i] is None:
                        continue
                    len_list = [len(x[inputs[i]["name"]]) for x in graphs_batch]
                    max_length = sum(len_list)
                    max_size_epoch[i].append(max_length)
            for i, x in max_size_epoch.items():
                if x is not None:
                    max_size[i].append(max(x))
        max_size = {i: max(x) if x is not None else None for i, x in max_size.items()}

        module_logger.info("Max of graph: %s." % total_max)
        module_logger.info("Padded max of disjoint: %s." % [
            x/batch_size if x is not None else None for x in max_size.values()])

    data_index = np.arange(dataset_size)
    rng = Generator(PCG64(seed=seed))

    def generator():

        if shuffle:
            rng.shuffle(data_index)

        for batch_index in range(0, dataset_size, batch_size):
            idx = data_index[batch_index:batch_index + batch_size]
            graphs_batch = [graphs[i] for i in idx]

            out = {i: None for i in inputs.keys()}
            out_counts = {i: None for i in inputs.keys()}

            for i in inputs.keys():
                if not is_attributes[i]:
                    continue

                array_list = [x[inputs[i]["name"]] for x in graphs_batch]
                if assignment_to_id[i] is None:
                    values = np.array(array_list, dtype=inputs[i]["dtype"])
                    if padded_disjoint:
                        values = pad_at_axis(values, (1, 0), axis=0)
                    out[i] = values
                else:
                    values = np.concatenate(array_list, axis=0)
                    counts = np.array([len(x) for x in array_list], dtype="int64")
                    ids = assignment_to_id[i]

                    if not padded_disjoint:
                        out[i] = values
                        out_counts[i] = counts
                    else:
                        len_values = len(values)
                        num_pad_required = max_size[i] - len_values + 1
                        values = pad_at_axis(values, (num_pad_required, 0), axis=0)
                        out[i] = values
                        counts = np.concatenate([np.array([num_pad_required], dtype=counts.dtype), counts], axis=0)
                        out_counts[i] = counts

                    if ids in pos_count:
                        if out[pos_count[ids]] is None:
                            out[pos_count[ids]] = counts
                    if ids in pos_batch_id:
                        if out[pos_batch_id[ids]] is None:
                            out[pos_batch_id[ids]] = np.repeat(
                                np.arange(len(counts), dtype="int64"), repeats=counts)
                    if ids in pos_subgraph_id:
                        if out[pos_subgraph_id[ids]] is None:
                            out[pos_subgraph_id[ids]] = np.concatenate(
                                [np.arange(x, dtype="int64") for x in counts], axis=0)

            # Indices
            for i in inputs.keys():
                if assignment_of_indices[i] is not None:
                    edge_indices_flatten = out[i]
                    count_nodes = out_counts[assignment_of_indices[i]]
                    count_edges = out_counts[i]
                    node_splits = np.pad(np.cumsum(count_nodes), [[1, 0]])
                    offset_edge_indices = np.expand_dims(np.repeat(node_splits[:-1], count_edges), axis=-1)
                    disjoint_indices = edge_indices_flatten + offset_edge_indices
                    disjoint_indices = np.transpose(disjoint_indices)
                    out[i] = disjoint_indices

            # Match output container
            if is_list_input:
                out = tuple([out[i] for i in range(num_inputs)])
            if is_single_input:
                out = out[0]

            yield out

    data_loader = tf.data.Dataset.from_generator(
        generator,
        output_signature=output_spec
    )

    return data_loader



from kgcnn.layers.modules import Embedding
from kgcnn.layers.gather import GatherState, GatherNodesIngoing, GatherNodesOutgoing
from keras.layers import Concatenate, Dense, Flatten
from kgcnn.layers.aggr import AggregateLocalEdges
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.set2set import PoolingSet2SetEncoder
from kgcnn.layers.mlp import MLP, GraphMLP


def model_disjoint(
        inputs,
        use_node_embedding,
        use_edge_embedding,
        use_graph_embedding,
        input_node_embedding,
        input_edge_embedding,
        input_graph_embedding,
        gather_args,
        depth,
        edge_mlp_args,
        pooling_args,
        node_mlp_args,
        output_embedding,
        use_set2set,
        set2set_args,
        output_mlp):
    # Make input
    n, ed, edi, uenv, batch_id_node, batch_id_edge, count_nodes, count_edges = inputs

    if use_node_embedding:
        n = Embedding(**input_node_embedding)(n)
    if use_edge_embedding:
        ed = Embedding(**input_edge_embedding)(ed)
    if use_graph_embedding:
        uenv = Embedding(**input_graph_embedding)(uenv)

    # Model
    ev = GatherState(**gather_args)([uenv, batch_id_node])
    # n-Layer Step
    for i in range(0, depth):
        # upd = GatherNodes()([n,edi])
        eu1 = GatherNodesIngoing(**gather_args)([n, edi])
        eu2 = GatherNodesOutgoing(**gather_args)([n, edi])
        upd = Concatenate(axis=-1)([eu2, eu1])
        eu = Concatenate(axis=-1)([upd, ed])

        eu = GraphMLP(**edge_mlp_args)([eu, batch_id_edge, count_edges])
        # Pool message
        nu = AggregateLocalEdges(**pooling_args)([n, eu, edi])  # Summing for each node connection
        # Add environment
        nu = Concatenate(axis=-1)(
            [n, nu, ev])  # LazyConcatenate node features with new edge updates
        n = GraphMLP(**node_mlp_args)([nu, batch_id_node, count_nodes])

    # Output embedding choice
    if output_embedding == 'graph':
        if use_set2set:
            # output
            n = Dense(set2set_args["channels"], activation="linear")(n)
            out = PoolingSet2SetEncoder(**set2set_args)([count_nodes, n, batch_id_node])
        else:
            out = PoolingNodes(**pooling_args)([count_nodes, n, batch_id_node])
        out = Flatten()(out)
        out = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = GraphMLP(**output_mlp)([n, batch_id_node, count_nodes])
    else:
        raise ValueError("Unsupported output embedding for mode `INorp` .")

    return out
